import subprocess
import unittest
import tempfile
import shutil

from pathlib import Path
from unittest import TestCase, mock

import jinja2

from bigflow.cli import cli
from bigflow.scaffold import templating as tt
from bigflow.scaffold import infra

from test import mixins


class ProjectScaffoldE2E(
    mixins.FileUtilsMixin,
    mixins.TempCwdMixin,
    unittest.TestCase,
):
    maxDiff = 5000

    @mock.patch('bigflow.cli.gcloud_project_list')
    @mock.patch('bigflow.cli.gcp_bucket_input')
    @mock.patch('bigflow.cli.gcp_project_input')
    @mock.patch('bigflow.cli.project_type_input')
    @mock.patch('bigflow.cli.project_name_input')
    @mock.patch('bigflow.build.pip.pip_compile')
    def test_should_create_basic_project(
        self,
        pip_compile_mock,
        project_name_input_mock,
        project_type_input_mock,
        gcp_project_input_mock,
        gcp_bucket_input_mock,
        gcloud_project_list_mock,
    ):
        # given
        gcloud_project_list_mock.return_value = ''
        project_name_input_mock.return_value = 'my_project'
        project_type_input_mock.return_value = 'b'
        gcp_project_input_mock.return_value = 'my_gcp_project'
        gcp_bucket_input_mock.return_value = 'my_gcp_bucket'

        # when
        cli(['start-project'])

        # then
        self.check_scaffolded_basic_project()

        # and
        self.scaffolded_project_tests_should_work()

    @unittest.skip('advanced mode not available yet')
    @mock.patch('bigflow.cli.project_number_input')
    @mock.patch('bigflow.cli.environment_name_input')
    @mock.patch('bigflow.cli.gcp_bucket_input')
    @mock.patch('bigflow.cli.gcp_project_input')
    @mock.patch('bigflow.cli.project_type_input')
    @mock.patch('bigflow.cli.project_name_input')
    def test_should_create_advanced_project(self, project_name_input_mock, project_type_input_mock,
                                            gcp_project_input_mock, gcp_bucket_input_mock, environment_name_input,
                                            project_number_input_mock):
        # given
        project_name_input_mock.return_value = 'my_project'
        project_type_input_mock.return_value = 'a'
        gcp_project_input_mock.side_effect = ['dev_project', 'test_project', 'prod_project']
        gcp_bucket_input_mock.side_effect = ['dev_bucket', 'test_bucket', 'prod_bucket']
        environment_name_input.side_effect = ['DEV', 'TEST', 'PROD']
        project_number_input_mock.return_value = '3'

        # when
        cli(['start-project'])

        # then
        self.scaffolded_advanced_project_should_have_three_environments()

        # and
        self.scaffolded_project_tests_should_work()

    def scaffolded_project_tests_should_work(self):
        cwd = Path('my_project_project')
        res = subprocess.run("python -m unittest discover -s test -p '*.py'", cwd=cwd, check=True, capture_output=True, shell=True)
        self.assertRegexpMatches(res.stderr.decode(), r"OK$")

    def check_scaffolded_basic_project(self):

        self.assertFileExists("my_project_project/my_project/wordcount/pipeline.py")
        self.assertFileContentRegex(
            "my_project_project/deployment_config.py",
            r"""deployment_config = DeploymentConfig\(""")

    def scaffolded_advanced_project_should_have_three_environments(self):
        self.check_file_content(Path('my_project_project') / 'deployment_config.py', '''from bigflow.configuration import Config

deployment_config = Config(name='dev',
                           properties={
                               'docker_repository': 'test_repository',
                               'gcp_project_id': 'dev_project',
                               'dags_bucket': 'dev_bucket'
                           }).add_configuration(name='TEST',
                           properties={
                               'gcp_project_id': 'test_project',
                               'dags_bucket': 'test_bucket'
                           }).add_configuration(name='PROD',
                           properties={
                               'gcp_project_id': 'prod_project',
                               'dags_bucket': 'prod_bucket'
                           })
''')
        self.check_file_content(Path('my_project_project') / 'workflows' / 'internationalports' / 'workflow.py', """from bigflow import Workflow
from bigflow.bigquery import DatasetConfig

dataset_config = DatasetConfig(
    env='dev',
    project_id='dev_project',
    dataset_name='internationalports',
    internal_tables=['ports', 'polish_ports'],
    external_tables={})

dataset = dataset_config.create_dataset_manager()

create_polish_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS polish_ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64)
''')

create_ports_table = dataset.create_table('''
    CREATE TABLE IF NOT EXISTS ports (
      port_name STRING,
      port_latitude FLOAT64,
      port_longitude FLOAT64,
      country STRING,
      index_number STRING)
''')

select_polish_ports = dataset.write_truncate('ports', '''
        SELECT port_name, port_latitude, port_longitude
        FROM `{more_ports}`
        WHERE country = 'POL'
        ''', partitioned=False)

populate_ports_table = dataset.collect('''
        INSERT INTO `{more_ports}` (port_name, port_latitude, port_longitude, country, index_number)
        VALUES
        ('GDYNIA', 54.533333, 18.55, 'POL', '28740'),
        ('GDANSK', 54.35, 18.666667, 'POL', '28710'),
        ('MURMANSK', 68.983333, 33.05, 'RUS', '62950'),
        ('SANKT-PETERBURG', 59.933333, 30.3, 'RUS', '28370');
        ''')


internationalports_workflow = Workflow(
        workflow_id='internationalports',
        definition=[
                create_ports_table.to_job(id='create_ports_table'),
                create_polish_ports_table.to_job(id='create_polish_ports_table'),
                populate_ports_table.to_job(id='populate_ports_table'),
                select_polish_ports.to_job(id='select_polish_ports'),
        ],
        schedule_interval='@once')""")

    def check_file_content(self, path, template):
        with open(path.resolve(), 'r') as f:
            self.assertEqual(f.read(), template)


class TemplatingTestCase(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def readfile(self, path):
        p = Path(self.tmp_dir) / Path(path)
        self.assertTrue(p.exists())
        return p.read_text()

    def test_basic_generate_templates(self):
        # given
        loader = jinja2.DictLoader({
            "dir1/dir2/template1.txt.jinja": "{% if a %}A{% else %}B{% endif %}",
            "dir1/dir2/template2.txt.j2": "{{ a }}",
            "dir1/dir2/template3.txt": "{{ a | title }}",
        })

        # when
        tt.render_templates(self.tmp_dir, loader, {'a': "value"})

        # when
        self.assertEqual(self.readfile("dir1/dir2/template1.txt"), "A")
        self.assertEqual(self.readfile("dir1/dir2/template2.txt"), "value")
        self.assertEqual(self.readfile("dir1/dir2/template3.txt"), "Value")

    def test_substitute_filenames(self):
        # given
        loader = jinja2.DictLoader({
            "{{one}}/{{two}}.txt": "OK",
            "{{10 * 20}}.txt": "OK",
            "a/{{long_path}}/b": "OK",
            "filename.{{ext}}": "OK",
        })

        # when
        tt.render_templates(
            self.tmp_dir, loader, {
                'one': 1,
                'two': 2,
                'long_path': "x/y/z",
                'ext': "xyz",
            })

        # then
        for expected_file in [
            "1/2.txt",
            "200.txt",
            "a/x/y/z/b",
            "filename.xyz",
        ]:
            self.assertEqual(self.readfile(expected_file), "OK")

    def test_conditional_render(self):
        # given
        loader = jinja2.DictLoader({
            "a1": "{% skip_file_when a %}X",
            "a2": "{% skip_file_unless a %}OK",
            "b1": "{% skip_file_when b %}OK",
            "b2": "{% skip_file_unless b %}X",
            "c1": "{% skip_file_when c %}OK",
            "c2": "{% skip_file_unless c %}X",
        })

        # when
        tt.render_templates(
            self.tmp_dir, loader, {
                'a': True,
                'b': False,
                'c': False,
            })

        # then
        for expected_file in [
            "a2", "b1", "c1",
        ]:
            self.assertEqual(self.readfile(expected_file), "OK")

        for should_not_exist in [
            "a1", "b2", "c2",
        ]:
            self.assertFalse((Path(self.tmp_dir) / should_not_exist).exists())


class CloudComposerCreatorTestCase(unittest.TestCase):

    infrastructure_parameters = (
        'fake-gcp-project',
        'fake-bigflow-project',
        'dev',
        'europe-west2',
        'europe-west2-b'
    )

    @mock.patch('bigflow.scaffold.infra.run_process')
    def test_should_create_cloud_composer(self, run_process_mock):
        # given
        gcp_project_id, bigflow_project_name, environment_name, region, zone = self.infrastructure_parameters

        # when
        infra.try_create(infra.CloudComposer(
            gcp_project_id,
            bigflow_project_name,
            environment_name,
            region,
            zone
        ))

        # then
        expected_router_name = f'bigflow-router-{bigflow_project_name}-{environment_name}-{region}'
        run_process_mock.assert_has_calls([
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'create', expected_router_name,
                '--network', 'default',
                '--region', region
            ]),
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'nats', 'create', f'bigflow-nat-{bigflow_project_name}-{environment_name}-{region}',
                f'--router={expected_router_name}',
                '--auto-allocate-nat-external-ips',
                '--nat-all-subnet-ip-ranges',
                '--enable-logging'
            ]),
            mock.call([
                'gcloud', 'composer',
                '--project', gcp_project_id,
                'environments', 'create', f'bigflow-composer-{bigflow_project_name}-{environment_name}-{region}',
                f'--location={region}',
                f'--zone={zone}',
                f'--env-variables=env={environment_name}',
                f'--machine-type=n1-standard-2',
                f'--node-count=3',
                f'--python-version=3',
                f'--enable-ip-alias',
                f'--network=default',
                f'--subnetwork=default',
                f'--enable-private-environment',
                f'--enable-private-endpoint',
            ])
        ])

    @mock.patch('bigflow.scaffold.infra.run_process')
    @mock.patch('bigflow.scaffold.infra._composer_create_command')
    def test_should_destroy_leftovers_when_error_occurs_during_creation(self, composer_create_command_mock, run_process_mock):
        # given
        def raise_error(*args):
            raise ValueError("Fail creation of composer command")

        gcp_project_id, bigflow_project_name, environment_name, region, zone = self.infrastructure_parameters
        composer_create_command_mock.side_effect = raise_error

        # when
        infra.try_create(infra.CloudComposer(
            gcp_project_id,
            bigflow_project_name,
            environment_name,
            region,
            zone
        ))

        # then
        expected_router_name = f'bigflow-router-{bigflow_project_name}-{environment_name}-{region}'
        expected_composer_name = f'bigflow-composer-{bigflow_project_name}-{environment_name}-{region}'
        expected_nat_name = f'bigflow-nat-{bigflow_project_name}-{environment_name}-{region}'

        run_process_mock.assert_has_calls([
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'create', expected_router_name,
                '--network', 'default',
                '--region', region
            ]),
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'nats', 'create', f'bigflow-nat-{bigflow_project_name}-{environment_name}-{region}',
                f'--router={expected_router_name}',
                '--auto-allocate-nat-external-ips',
                '--nat-all-subnet-ip-ranges',
                '--enable-logging'
            ]),
            mock.call([
                'gcloud', 'composer',
                'environments', 'delete', expected_composer_name,
                '--location', region,
                '--project', gcp_project_id,
                '--quiet'
            ]),
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'nats', 'delete', expected_nat_name,
                '--router', expected_router_name,
                '--quiet'
            ]),
            mock.call([
                'gcloud', 'compute',
                '--project', gcp_project_id,
                'routers', 'delete', expected_router_name,
                '--region', region,
                '--quiet'
            ])
        ])


import bigflow as bf


workflow_1 = bf.Workflow(workflow_id="ID_1", definition=[], schedule_interval="@once", log_config={
        'gcp_project_id': 'some-project-id',
        'log_level': 'INFO',
    })
workflow_2 = bf.Workflow(workflow_id="ID_2", definition=[], log_config={
        'gcp_project_id': 'some-project-id',
        'log_level': 'INFO',
    })


"""Support for 'self-builded' packages.

Provide access to embedded sdist distribution,
allowing to build whl/egg/tar packages from pip-installed source.

Thoose packages may be latter used by pyspark, beam, etc.
"""


import sys
import os
import inspect
import logging
import tarfile
import tempfile
import textwrap
import types
import pkg_resources
import importlib
import functools

from typing import Optional, List, Tuple, Union
from pathlib import Path

import bigflow.commons as bfc
import bigflow.build.dev
import bigflow.build.spec

from bigflow.build.spec import BigflowProjectSpec
from bigflow.commons import public


logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=None)
def _iter_dist_toplevel_packages(distname) -> List[str]:
    try:
        dist = pkg_resources.get_distribution(distname)
        result = list(dist.get_metadata_lines("top_level.txt"))
        logger.debug("Distribution %r has packages %s", dist, result)
        return result
    except FileNotFoundError:
        logger.debug("Distribution %r doesn't have 'top_level.txt' metadata - skip", distname)
        return []
    except pkg_resources.DistributionNotFound:
        logger.error("Unknown distribution %r", distname)
        return []


@functools.lru_cache(maxsize=None)
def _get_topmodule_to_dist_mapping():
    all_dist_names = list(pkg_resources.Environment())
    return {
        p: dname
        for dname in all_dist_names
        for p in _iter_dist_toplevel_packages(dname)
    }


def _module_to_enclosing_directory(module: types.ModuleType) -> Path:
    if getattr(module, '__path__', None):
        return Path(next(iter(module.__path__)))
    if module.__file__:
        file = Path(module.__file__)
        return file.parent


def _infer_project_name_by_distribution(module: types.ModuleType) -> Optional[str]:
    top = module.__name__.split(".", 2)[0]
    top_to_dist_name = _get_topmodule_to_dist_mapping()
    try:
        dist_name = top_to_dist_name[top]
    except KeyError:
        return None
    else:
        dist = pkg_resources.get_distribution(dist_name)
        return dist.project_name


def _locate_dev_project_directory_by_module(module: types.ModuleType) -> Optional[Path]:
    """Tries to locate setup.py/pyproject.toml near the module enclosing folder"""

    top_name = module.__name__.split(".", 2)[0]
    top_module = importlib.import_module(top_name)
    d = _module_to_enclosing_directory(top_module)
    d = d.resolve()

    try:
        return bigflow.build.dev.find_project_dir(d)
    except FileNotFoundError:
        return False


@public()
def locate_project_path(project_name: Union[str, Path, None]) -> Path:
    """Returns path to either project sdist package (tar) or directory with project sources"""

    if isinstance(project_name, Path):
        logger.debug("Given explicit project location - do nothing")
        return project_name

    elif project_name:
        logger.debug("Explicit project name - try to locate self-installed sdist package")
        pkg = _locate_self_package(project_name)
        if pkg:
            logger.debug("Self-package was found at %s", pkg)
            return pkg
        logger.debug("No self-package found - maybe our project is not pip-installed")

    # no explicit package name - inspect stack to find setup.py / pyproject.toml or detect project name
    logger.debug("Inspect stack to find project root directory...")

    s = inspect.stack()
    for frame_info in s:
        logger.debug("Analyze frame %s", frame_info)

        m = inspect.getmodule(frame_info[0])
        if not m:
            continue
        logger.debug("Module is %s - %s", m, m.__name__)

        if m.__name__.startswith("bigflow."):
            continue

        pdir = _locate_dev_project_directory_by_module(m)
        if pdir:
            logger.debug("Found project sources at %s", pdir)
            return pdir

        pname = _infer_project_name_by_distribution(m)
        if pname:
            logger.debug("Project name inferred to %s", pname)
            pkg = _locate_self_package(pname)
            if pkg:
                return pkg

    raise RuntimeError("Unable to infer project path")


def _locate_self_package(project_name) -> Optional[Path]:
    p = Path(sys.prefix) / "bigflow__project" / project_name / "bf-project.tar"
    if p.exists():
        logger.debug("Found sdist distribution for project %r: %s", project_name, p)
        return p
    else:
        logger.debug("Sdist distribution for project %r not found", project_name)
        return None


def _expect_single_file(directory: str, pattern: str) -> Path:
    logger.debug("Expect that directory %s contains only one file %s", directory, pattern)
    directory = Path(directory)
    fs = list(directory.glob(pattern))
    if not fs:
        raise FileNotFoundError(f"no files matching {pattern} at {directory}")
    elif len(fs) == 1:
        logger.debug("Found file %s", fs[0])
        return fs[0]
    else:
        logger.warning("Found multiple files at %s, use the first one: %s", directory, fs)
        return fs[0]


@public()
def get_project_spec(
    project_name: Union[str, Path, None] = None,
) -> BigflowProjectSpec:
    """Read & parse current project spec.

    This function should be called from inside the project as
    it uses dynamic inspection to find project name.
    """

    pkg = locate_project_path(project_name)
    logger.debug("Project loated at %s", pkg)

    if pkg.is_dir():
        pdir = pkg
    else:
        # sdist-package & pyproject.toml are located in the same directory
        pdir = pkg.parent
        assert (pdir / "pyproject.toml").exists()

    return bigflow.build.spec.read_project_spec(pdir)


def _ensure_setuppy_exists(setuppy: Path):
    if setuppy.exists():
        return
    elif (setuppy.parent / "pyproject.toml").exists():
        logger.debug("Write dummy setup.py at %s", setuppy)
        setuppy.write_text(textwrap.dedent("""
            # dummy setup.py, generated by 'bigflow', see pyproject.toml for project parameters
            import bigflow.build; buigflow.build.setup()
        """))
    else:
        raise FileNotFoundError("Unable to find file", setuppy)


@public()
def materialize_setuppy(
    project_name: Optional[Union[str, Path]] = None,
    tempdir: Optional[str] = None,
) -> Path:
    """Locates project setup.py.  Unpacks embedded sdist distribution when needed."""

    pkg = locate_project_path(project_name)

    if pkg.is_dir():
        setuppy = pkg / "setup.py"
        _ensure_setuppy_exists(setuppy)
        return setuppy
    else:
        tempdir = tempdir or tempfile.mkdtemp()
        logger.info("Unpack %s to %s", pkg, tempdir)
        tarfile.open(pkg).extractall(tempdir)
        pkgdir = _expect_single_file(tempdir, "*")
        setuppy = pkgdir / "setup.py"
        _ensure_setuppy_exists(setuppy)
        return setuppy


def _build_dist_package(
    project_name: str,
    suffix: str,
    cmdname: str,
    exargs: List[str],
):
    """Locates and runs 'bdist_*' command on 'setup.py'"""

    with tempfile.TemporaryDirectory() as workdir:

        setuppy = materialize_setuppy(project_name, workdir)
        distdir = Path(workdir) / "dist"
        logger.info("Run setup.py %s", cmdname)

        fd, result_path = tempfile.mkstemp(suffix=suffix)
        result_path = Path(result_path)
        os.close(fd)

        bfc.run_process(
            [
                "python", setuppy, cmdname,
                "--dist-dir", distdir,
                *(exargs or []),
            ],
            cwd=str(setuppy.parent),
        )
        result_tmp = _expect_single_file(distdir, "*" + suffix)

        logger.debug("Rename %s to %s", result_tmp, result_path)
        result_tmp.rename(result_path)

        logger.info("Built package located at %s", result_path)
        return result_path


@public()
def build_sdist(project_name=None) -> Path:
    """Builds project 'sdist' package"""
    return _build_dist_package(project_name, ".tar.gz", "sdist", ["--format", "gztar"])


@public()
def build_wheel(project_name=None) -> Path:
    """Builds project 'wheel' package"""
    return _build_dist_package(project_name, ".whl", "bdist_wheel", ["--compression", "deflated"])


@public()
def build_egg(project_name=None) -> Path:
    """Build project 'egg' package"""
    return _build_dist_package(project_name, ".egg", "bdist_egg", [])



import bigflow as bf
from bigflow.bigquery.interactive import InteractiveDatasetManager as Dataset

PROJECT_ID = 'put-you-project-id-here'

dataset = Dataset(
    project_id=PROJECT_ID,
    dataset_name='bigflow_cheatsheet',
    external_tables={
        '311_requests': '{}.external_data.311_requests'.format(PROJECT_ID)
    },
    internal_tables=['request_aggregate'])

wait_for_requests = bf.bigquery.sensor(
    '311_requests',
    where_clause="DATE(TIMESTAMP(created_date)) = DATE(TIMESTAMP_ADD(TIMESTAMP('{dt}'), INTERVAL -24 HOUR))",
    ds=dataset)

started_jobs = []

class ExampleJob:
    def __init__(self, id):
        self.id = id

    def execute(self, context):
        started_jobs.append(self.id)

workflow_1 = bf.Workflow(workflow_id="ID_1", definition=[wait_for_requests.to_job(), wait_for_requests.to_job()], schedule_interval="@once", log_config={
        'gcp_project_id': 'some-project-id',
        'log_level': 'INFO',
    })
workflow_2 = bf.Workflow(workflow_id="ID_2", definition=[wait_for_requests.to_job()], log_config={
        'gcp_project_id': 'another-project-id',
        'log_level': 'INFO',
    })
workflow_3 = bf.Workflow(workflow_id="ID_3", definition=[ExampleJob("J_ID_3"), ExampleJob("J_ID_4")])
workflow_4 = bf.Workflow(workflow_id="ID_4", definition=[ExampleJob("J_ID_5")])

print("AAA")



import re
import sys
import logging
import textwrap
import importlib.resources

from pathlib import Path
from typing import Dict, List

import bigflow.build.pip
import bigflow.build.spec
import bigflow.build.dataflow.workerdeps


logger = logging.getLogger(__name__)


def detect_py_version():
    vi = sys.version_info
    return f"{vi.major}.{vi.minor}"


def detect_beam_version(reqs: List[str]):
    reqs = build_requirements_dict(reqs)
    bv = reqs.get('apache-beam')
    return ".".join(bv.split(".")[:2]) if bv else None


def build_requirements_dict(requirements: List[str]) -> Dict[str, str]:
    matcher = re.compile(r"(.*?)(?:\[.*?\])?==(.*)")
    res = {}
    for line in requirements:
        m = matcher.match(line.lower())
        if m:
            k, v = m.groups()
            res[k] = v
    return res


def load_beam_worker_preinstalled_dependencies(beam_version, py_version):
    fname = f"beam{beam_version}_py{py_version}.txt"
    txt = importlib.resources.read_text(bigflow.build.dataflow.workerdeps, fname)
    result = dict(
        line.lower().split("==", 2)
        for line in txt.splitlines()
        if line.strip() and not line.startswith("#")
    )

    if result['google-cloud-core'] == "1.1.0":
        # Bumpup version of `google-cloud-storage` to 1.4.1
        # Google has preinstalled 1.1, but semi-fresh versions of google-cloud-storage depends on 1.4
        # Version 1.4 doesn't have any breaking changes since 1.1, so this upgrade should be safe.
        result['google-cloud-core'] = "1.4.1"

    return result


def detect_dataflow_conflicts(req_path: Path, all=False):

    req_path = req_path.with_suffix(".txt")
    existing_pinfile = req_path.parent / "dataflow_pins.in"

    requirements = bigflow.build.pip.read_requirements(req_path)
    beam_version = detect_beam_version(requirements)
    if not beam_version:
        logger.debug("Beam is not used - don't perform worker dependencies conflict check")
        return {}

    py_version = detect_py_version()
    existing_pins = existing_pinfile.read_text() if existing_pinfile.exists() else ""

    logger.debug("Beam version is %s", beam_version)
    logger.debug("Python version is %s", py_version)

    try:
        workerdeps = load_beam_worker_preinstalled_dependencies(beam_version, py_version)
    except FileNotFoundError:
        logger.error("Unsupported beam/python version: %s/%s", beam_version, py_version)
        return {}

    reqs = build_requirements_dict(requirements)
    common_deps = set(reqs) & set(workerdeps)
    conflicts = {
        k: (reqs[k], workerdeps[k])
        for k in reqs  # preserve order of `reqs`
        if k in common_deps
        if (reqs[k] != workerdeps[k] or all)
        and not re.search(rf"\W{re.escape(k)}\W", existing_pins)  # pin is ignored
    }

    return conflicts


def check_beam_worker_dependencies_conflict(req_path: Path):
    conflicts = detect_dataflow_conflicts(req_path)
    if not conflicts:
        logger.debug("No conflicts found - woow!")
    else:
        logger.error(
            textwrap.dedent(f"""
                ========================================
                The Apache Beam SDKs and Dataflow workers depend on common third-party components which then import additional dependencies.
                Your project is using different verions of those libraries, which may result in conflicts, bugs or unexpected behaviour.
                More information https://cloud.google.com/dataflow/docs/concepts/sdk-worker-dependencies
                Conflicted libraries:
                %s

                It is recommended to resolve conflicts by running:
                > bigflow codegen pin-dataflow-requirements
                ========================================
            """),
            "\n".join(
                f" - {dep}: used {our_version}, beam has {worker_version}"
                for dep, (our_version, worker_version) in conflicts.items()
            ),
        )


def sync_requirements_with_dataflow_workers(req_path=None):

    if req_path is None:
        params = bigflow.build.spec.get_project_spec()
        req_path = Path(params.project_requirements_file or "resources/requirements.txt")

    pins_in = req_path.parent / "dataflow_pins.in"
    return bigflow.build.pip.generate_pinfile(
        req_path,
        pins_in,
        lambda: [f"{a}=={c}" for a, (b, c) in detect_dataflow_conflicts(req_path, True).items()],
    )



from bigflow.build.operate import build_project
from unittest import TestCase
import itertools
import mock
import shutil
import freezegun

from bigflow.testing.isolate import ForkIsolateMixin
from bigflow.cli import *

from test import mixins

TESTS_DIR = Path(__file__).parent


class CliTestCase(
    mixins.PrototypedDirMixin,
    ForkIsolateMixin,
    TestCase,
):
    proto_dir = "bf-projects/example_project"

    def setUp(self) -> None:
        super().setUp()
        sys.path.append(str(Path(__file__).parent))

        self.project_setuppy = self.cwd / "setup.py"

        global TEST_MODULE_PATH
        TEST_MODULE_PATH = Path(__file__).parent / 'test_module'

        bigflow.build.spec.get_project_spec.cache_clear()

    def test_should_walk_through_all_files_inside_package_tree(self):
        # when
        res = walk_module_files(TEST_MODULE_PATH)

        # then all files are listed
        res_as_list = list(res)
        absolute = str(Path(TEST_MODULE_PATH).absolute())
        expected = [(absolute, "__init__.py"), (absolute, "Unused1.py"), (absolute, "Unused2.py"),
                    (absolute, "Unused3.py")]
        self.assertCountEqual(expected, res_as_list)

        # and at the first position there are absolute paths
        for (path, name) in res_as_list:
            self.assertEqual('/', path[0], "Path should be absolute and start with /")
            expected_ending = 'bigflow/test/cli/test_module'
            self.assertEqual(expected_ending, path[-len(expected_ending):])

    def test_should_walk_through_all_module_paths_inside_package_tree(self):
        # when
        res = walk_module_paths(TEST_MODULE_PATH)

        # then all packages are listed
        res_as_list = list(res)
        expected = ['test_module', 'test_module.Unused1', 'test_module.Unused2', 'test_module.Unused3']
        self.assertCountEqual(expected, list(res_as_list))

    def test_should_build_module_path_for_example_file(self):
        # given
        root_path = TEST_MODULE_PATH.parent
        module_file = "py_unused1.py"
        file_path = TEST_MODULE_PATH

        # when
        res = build_module_path(root_path, file_path, module_file)

        # then
        self.assertEqual("cli.test_module.py_unused1", res)

    def test_should_walk_through_all_modules_inside_package_tree(self):
        # when
        res = walk_modules(TEST_MODULE_PATH)

        # then
        res = list(res)
        self.assertEqual(4, len(res))
        expected = ['test_module', 'test_module.Unused1', 'test_module.Unused2', 'test_module.Unused3']
        self.assertCountEqual(expected, [x.__name__ for x in res])

        unused2 = sorted(res, key=lambda x: x.__name__)[2]
        self.assertIn('workflow_1', dir(unused2))
        self.assertNotIn('workflow_2', dir(unused2))

    def test_should_walk_through_all_module_objects_inside_package_tree(self):
        # given
        modules = sorted(list(walk_modules(TEST_MODULE_PATH)), key=lambda x: x.__name__)
        unused2 = modules[2]  # !!!

        # when looks in unused2
        res = walk_module_objects(unused2, bf.Workflow)

        # then finds 1 bf.Workflow objects
        res = list(res)
        self.assertEqual(1, len(res))

        # when looks in unused2
        res = walk_module_objects(unused2, int)

        # then finds 3 int objects
        res = list(res)
        self.assertEqual(3, len(res))

    def test_should_walk_through_all_workflows_inside_package_tree(self):
        # when
        res = walk_workflows(TEST_MODULE_PATH)

        # then
        res = sorted(res, key=lambda r: r.workflow_id)

        self.assertEqual(5, len(res))
        self.assertEqual('ID_1', res[0].workflow_id)
        self.assertEqual('@once', res[0].schedule_interval)
        self.assertEqual('ID_2', res[1].workflow_id)
        self.assertNotEqual('@once', res[1].schedule_interval)

    def test_should_find_existing_workflow(self):
        # when
        res = find_workflow(TEST_MODULE_PATH, 'ID_1')

        # then
        self.assertEqual(bf.Workflow, type(res))
        self.assertEqual('ID_1', res.workflow_id)
        self.assertEqual('@once', res.schedule_interval)

    def test_should_not_find_non_existing_workflow(self):
        with self.assertRaises(ValueError) as cm:
            # when
            find_workflow(TEST_MODULE_PATH, 'NOT_EXISTING_ID')

        # then
        exception_message = cm.exception.args[0]
        expected_prefix = "Workflow with id NOT_EXISTING_ID not found in package "
        expected_suffix = "bigflow/test/cli/test_module"
        self.assertEqual(exception_message[:len(expected_prefix)], expected_prefix)
        self.assertEqual(exception_message[-len(expected_suffix):], expected_suffix)

    def test_should_raise_exception_when_no_jobid_and_no_workflow(self):
        # given
        root_package = TESTS_DIR / "test_module"

        with self.assertRaises(ValueError):
            # when
            cli_run(root_package)

    def test_should_raise_exception_when_job_id_incorrect(self):
        # given
        root_package = TESTS_DIR / "test_module"

        with self.assertRaises(ValueError):
            # when just job id
            cli_run(root_package, full_job_id="J_ID_1")

        with self.assertRaises(ValueError):
            # when just workflow id
            cli_run(root_package, full_job_id="ID_1")


    @mock.patch.dict('os.environ')
    def test_should_set_configuration_env(self):
        # given
        import os
        to_set = "Come fake config"
        self.assertNotEqual(to_set, os.environ.get('bf_env', None))

        # when
        set_configuration_env(to_set)

        # then
        self.assertEqual(to_set, os.environ.get('bf_env', None))

    def test_should_find_root_package_when_root_used_project_name_used(self):
        # given
        test_module_src = str(TEST_MODULE_PATH)

        # when
        res = find_root_package(test_module_src, "some_other_path")

        # then
        self.assertEqual(test_module_src, str(res))

    def test_should_run_workflow(self):
        # given
        root_package = TESTS_DIR / "test_module"

        # when
        cli_run(root_package, workflow_id="ID_3")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_4'])

        # when
        cli_run(root_package, workflow_id="ID_4")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_4', 'J_ID_5'])

    def test_should_run_workflow_multiple_times(self):
        # given
        root_package = TESTS_DIR / "test_module"

        # when
        cli_run(root_package, workflow_id="ID_3")
        cli_run(root_package, workflow_id="ID_3")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_4', 'J_ID_3', 'J_ID_4'])

    def test_should_run_job(self):
        # given
        root_package = TESTS_DIR / "test_module"

        # when
        cli_run(root_package, full_job_id="ID_3.J_ID_3")

        # then
        self.assert_started_jobs(['J_ID_3'])

        # when
        cli_run(root_package, full_job_id="ID_3.J_ID_4")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_4'])

        # when
        cli_run(root_package, full_job_id="ID_4.J_ID_5")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_4', 'J_ID_5'])

    def test_should_run_job_multiple_times(self):
        # given
        root_package = TESTS_DIR / "test_module"

        # when
        cli_run(root_package, full_job_id="ID_3.J_ID_3")
        cli_run(root_package, full_job_id="ID_3.J_ID_3")

        # then
        self.assert_started_jobs(['J_ID_3', 'J_ID_3'])

    def test_should_read_project_package_if_set(self):
        # given
        args = lambda: None
        expected = "ROOT_VALUE"
        args.project_package = expected

        # when
        res = read_project_package(args)

        # then
        self.assertEqual(expected, res)

    def test_should_return_None_when_no_root_property(self):
        # given
        args = lambda: None

        # when
        res = read_project_package(args)

        # then
        self.assertEqual(None, res)

    def assert_started_jobs(self, ids):
        self.assertEqual(ids, import_module("test_module.Unused1").started_jobs)

    def test_should_raise_error_when_deployment_config_is_needed_but_missing(self):
        with self.assertRaises(ValueError):
            # when
            cli(['deploy-dags'])

    @mock.patch('bigflow.cli.deploy_dags_folder')
    def test_should_call_cli_deploy_dags_command__with_defaults_and_with_implicit_deployment_config_file(self,
                                                                                                         deploy_dags_folder_mock):
        # given
        self._touch_file('deployment_config.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                           properties={
                               'gcp_project_id': 'my-gcp-project-id',
                               'dags_bucket': 'my-dags-bucket',
                               'vault_secret': 'secret'
                           })
        ''')

        # when
        cli(['deploy-dags'])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='local_account',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-dags-bucket',
                                                   dags_dir=self._expected_default_dags_dir(),
                                                   project_id='my-gcp-project-id',
                                                   vault_endpoint=None,
                                                   vault_secret='secret')

    @mock.patch('bigflow.cli.deploy_dags_folder')
    def test_should_call_cli_deploy_dags_command_for_different_environments(self, deploy_dags_folder_mock):
        # given
        self._touch_file('deployment_config.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                          properties={
                              'gcp_project_id': 'my-gcp-dev-project-id',
                              'dags_bucket': 'my-dags-dev-bucket',
                              'vault_secret': 'secret-dev'
                          })\
    .add_configuration(name='prod',
                          properties={
                              'gcp_project_id': 'my-gcp-prod-project-id',
                              'dags_bucket': 'my-dags-prod-bucket',
                              'vault_secret': 'secret-prod'
                          })

        ''')

        # when
        cli(['deploy-dags'])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='local_account',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-dags-dev-bucket',
                                                   dags_dir=self._expected_default_dags_dir(),
                                                   project_id='my-gcp-dev-project-id',
                                                   vault_endpoint=None,
                                                   vault_secret='secret-dev')

        # when
        cli(['deploy-dags', '--config', 'dev'])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='local_account',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-dags-dev-bucket',
                                                   dags_dir=self._expected_default_dags_dir(),
                                                   project_id='my-gcp-dev-project-id',
                                                   vault_endpoint=None,
                                                   vault_secret='secret-dev')

        # when
        cli(['deploy-dags', '--config', 'prod'])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='local_account',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-dags-prod-bucket',
                                                   dags_dir=self._expected_default_dags_dir(),
                                                   project_id='my-gcp-prod-project-id',
                                                   vault_endpoint=None,
                                                   vault_secret='secret-prod')

    @mock.patch('bigflow.cli.deploy_dags_folder')
    def test_should_call_cli_deploy_dags_command__when_parameters_are_given_by_explicit_deployment_config_file(self,
                                                                                                               deploy_dags_folder_mock):
        # given
        dc_file = self._touch_file('deployment_config_another.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                        properties={
                               'gcp_project_id': 'my-another-gcp-project-id',
                               'vault_endpoint': 'my-another-vault-endpoint',
                               'dags_bucket': 'my-another-dags-bucket',
                               'vault_secret': 'secrett'
                        })
        ''')

        # when
        cli(['deploy-dags',
             '--deployment-config-path', dc_file.as_posix(),
             '--dags-dir', '/tmp/my-dags-dir',
             '--auth-method', 'vault'
             ])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='vault',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-another-dags-bucket',
                                                   dags_dir='/tmp/my-dags-dir',
                                                   project_id='my-another-gcp-project-id',
                                                   vault_endpoint='my-another-vault-endpoint',
                                                   vault_secret='secrett')

    @mock.patch('bigflow.cli.deploy_dags_folder')
    def test_should_call_cli_deploy_dags_command__when_all_parameters_are_given_by_cli_arguments(self,
                                                                                                 deploy_dags_folder_mock):
        # when
        cli(['deploy-dags',
             '--dags-bucket', 'my-dags-bucket',
             '--dags-dir', '/tmp/my-dags-dir',
             '--vault-endpoint', 'my-vault-endpoint',
             '--gcp-project-id', 'my-gcp-project-id',
             '--auth-method', 'vault',
             '--clear-dags-folder',
             '--vault-secret', 'secrett'
             ])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='vault',
                                                   clear_dags_folder=True,
                                                   dags_bucket='my-dags-bucket',
                                                   dags_dir='/tmp/my-dags-dir',
                                                   project_id='my-gcp-project-id',
                                                   vault_endpoint='my-vault-endpoint',
                                                   vault_secret='secrett')

    @mock.patch('bigflow.cli.deploy_docker_image')
    def test_should_call_cli_deploy_image_command__with_defaults_and_with_implicit_deployment_config_file(self,
                                                                                                          deploy_docker_image_mock):
        # given
        self._touch_file('deployment_config.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                          properties={
                              'docker_repository': 'my-docker--repository'
                          })
        ''')

        # when
        cli(['deploy-image', '--image-tar-path', 'image-0.0.2.tar'])

        # then
        deploy_docker_image_mock.assert_called_with(auth_method='local_account',
                                                    docker_repository='my-docker--repository',
                                                    image_tar_path='image-0.0.2.tar',
                                                    vault_endpoint=None,
                                                    vault_secret=None)

    @mock.patch('bigflow.cli.deploy_docker_image')
    def test_should_call_cli_deploy_image_command__with_explicit_deployment_config_file(self, deploy_docker_image_mock):
        # given
        dc_file = self._touch_file('my_deployment_config.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                         properties={
                             'docker_repository': 'my-another-docker-repository' ,
                             'vault_endpoint' : 'my-another-vault-endpoint'
                         })
        ''')

        # when
        cli(['deploy-image',
             '--image-tar-path', 'image-0.0.3.tar',
             '--deployment-config-path', dc_file.as_posix(),
             '--auth-method', 'vault',
             '--vault-secret', 'secrett'
             ])

        # then
        deploy_docker_image_mock.assert_called_with(auth_method='vault',
                                                    docker_repository='my-another-docker-repository',
                                                    image_tar_path='image-0.0.3.tar',
                                                    vault_endpoint='my-another-vault-endpoint',
                                                    vault_secret='secrett')

    @mock.patch('bigflow.cli.deploy_docker_image')
    def test_should_call_cli_deploy_image_command__when_all_parameters_are_given_by_cli_arguments_and_image_is_loaded_from_tar(
            self, deploy_docker_image_mock):
        # when
        cli(['deploy-image',
             '--image-tar-path', 'image-0.0.1.tar',
             '--docker-repository', 'my-docker-repository',
             '--vault-endpoint', 'my-vault-endpoint',
             '--auth-method', 'vault',
             '--vault-secret', 'secrett'
             ])

        # then
        deploy_docker_image_mock.assert_called_with(auth_method='vault',
                                                    docker_repository='my-docker-repository',
                                                    image_tar_path='image-0.0.1.tar',
                                                    vault_endpoint='my-vault-endpoint',
                                                    vault_secret='secrett')

    @mock.patch('bigflow.cli.deploy_docker_image')
    def test_should_find_tar_in_image_directory(self, deploy_docker_image_mock):
        # given

        shutil.rmtree(Path.cwd() / ".image", ignore_errors=True)
        self._touch_file('image-123.tar', '', '.image')

        # when
        cli(['deploy-image',
             '--docker-repository', 'my-docker-repository',
             '--vault-endpoint', 'my-vault-endpoint',
             '--auth-method', 'vault',
             '--vault-secret', 'secrett'
             ])

        # then
        deploy_docker_image_mock.assert_called_with(auth_method='vault',
                                                    docker_repository='my-docker-repository',
                                                    image_tar_path='.image/image-123.tar',
                                                    vault_endpoint='my-vault-endpoint',
                                                    vault_secret='secrett')

    @mock.patch('bigflow.cli.deploy_dags_folder')
    @mock.patch('bigflow.cli.deploy_docker_image')
    def test_should_call_both_deploy_methods_with_deploy_command(self, deploy_docker_image_mock,
                                                                 deploy_dags_folder_mock):
        # given
        self._touch_file('deployment_config.py',
        '''
from bigflow import Config

deployment_config = Config(name='dev',
                         properties={
                             'docker_repository': 'my-docker--repository',
                             'gcp_project_id': 'my-gcp-project-id',
                            'dags_bucket': 'my-dags-bucket'
                         })
        ''')

        # when
        cli(['deploy', '-i', 'my-images/image-version'])

        # then
        deploy_dags_folder_mock.assert_called_with(auth_method='local_account',
                                                   clear_dags_folder=False,
                                                   dags_bucket='my-dags-bucket',
                                                   dags_dir=self._expected_default_dags_dir(),
                                                   project_id='my-gcp-project-id',
                                                   vault_endpoint=None,
                                                   vault_secret=None)

        deploy_docker_image_mock.assert_called_with(auth_method='local_account',
                                                    docker_repository='my-docker--repository',
                                                    image_tar_path='my-images/image-version',
                                                    vault_endpoint=None,
                                                    vault_secret=None)

    @mock.patch('bigflow.cli._cli_build_dags')
    def test_should_call_cli_build_dags_command(self, _cli_build_dags_mock):
        # when
        cli(['build-dags'])

        # then
        _cli_build_dags_mock.assert_called_with(Namespace(operation='build-dags', start_time=None, workflow=None, verbose=False))

        # when
        cli(['build-dags', '-t', '2020-01-01 00:00:00'])

        # then
        _cli_build_dags_mock.assert_called_with(Namespace(operation='build-dags', start_time='2020-01-01 00:00:00', workflow=None, verbose=False))

        # when
        cli(['build-dags', '-w', 'some_workflow'])

        # then
        _cli_build_dags_mock.assert_called_with(Namespace(operation='build-dags', start_time=None, workflow='some_workflow', verbose=False))

        # when
        cli(['build-dags', '-w', 'some_workflow', '-t', '2020-01-01 00:00:00'])

        # then
        _cli_build_dags_mock.assert_called_with(Namespace(operation='build-dags', start_time='2020-01-01 00:00:00', workflow='some_workflow', verbose=False))

        # when
        cli(['build-dags', '-w', 'some_workflow', '-t', '2020-01-01'])

        # then
        _cli_build_dags_mock.assert_called_with(Namespace(operation='build-dags', start_time='2020-01-01', workflow='some_workflow', verbose=False))

        # when
        with self.assertRaises(SystemExit):
            cli(['build-dags', '-w', 'some_workflow', '-t', '20200101'])

    @mock.patch('bigflow.build.operate.build_dags')
    @mock.patch('bigflow.build.spec.read_project_spec')
    @freezegun.freeze_time("2001-02-03 15:00:00")
    def test_should_call_cli_build_dags_commands_with_NOW_and_ALL(
        self,
        read_project_spec: mock.Mock,
        build_dags_mock: mock.Mock,
    ):
        # when
        cli(['build-dags','--workflow', 'ALL', '--start-time', 'NOW'])

        # then
        read_project_spec.assert_called_once()
        build_dags_mock.assert_any_call(
            read_project_spec.return_value,
            start_time="2001-02-03 15:00:00",
            workflow_id=None,
        )

    @mock.patch('bigflow.build.operate.build_project')
    @mock.patch('bigflow.build.spec.read_project_spec')
    @freezegun.freeze_time("2001-02-03 15:00:00")
    def test_should_call_cli_build_commands_with_NOW_and_ALL(
        self,
        read_project_spec: mock.Mock,
        build_project_mock: mock.Mock,
    ):
        # when
        cli(['build','--workflow', 'ALL', '--start-time', 'NOW'])

        # then
        read_project_spec.assert_called_once()
        build_project_mock.assert_any_call(
            read_project_spec.return_value,
            start_time="2001-02-03 15:00:00",
            workflow_id=None,
        )

    @mock.patch('bigflow.cli._cli_build_image')
    def test_should_call_cli_build_image_command(self, _cli_build_image_mock):
        # when
        cli(['build-image'])

        # then
        _cli_build_image_mock.assert_called_with(Namespace(operation='build-image', verbose=False))

    @mock.patch('bigflow.build.operate.build_project')
    @mock.patch('bigflow.build.spec.read_project_spec')
    @freezegun.freeze_time("2001-02-03 15:00:00")
    def test_should_call_build_command_through_CLI(
        self,
        read_project_spec: mock.Mock,
        build_project_mock: mock.Mock,
    ):
        # when
        cli(['build'])

        # then
        read_project_spec.assert_called_once()
        build_project_mock.assert_any_call(
            read_project_spec.return_value,
            start_time="2001-02-03 15:00:00",
            workflow_id=None,
        )

    @mock.patch('bigflow.cli._cli_build_package')
    def test_should_call_cli_build_package_command(self, _cli_build_package_mock):
        # when
        cli(['build-package'])

        # then
        _cli_build_package_mock.assert_called_with()

    @mock.patch('bigflow.cli._cli_build')
    def test_should_call_cli_build_command(self, _cli_build_mock):
        # when
        cli(['build'])

        # then
        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time=None, workflow=None, verbose=False))

        # when
        cli(['build', '--start-time', '2020-01-01 00:00:00'])

        # then

        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time='2020-01-01 00:00:00', workflow=None, verbose=False))

        # when
        cli(['build', '--start-time', '2020-01-01 00:00:00', '--workflow', 'some_workflow'])

        # then
        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time='2020-01-01 00:00:00', workflow='some_workflow', verbose=False))

    @mock.patch('bigflow.build.operate.build_package')
    @mock.patch('bigflow.build.spec.read_project_spec')
    def test_should_call_build_package_command_through_CLI(
        self,
        read_project_mock: mock.Mock,
        build_package_mock: mock.Mock,
    ):

        # when
        cli(['build-package'])

        # then
        read_project_mock.assert_called_once()
        build_package_mock.assert_any_call(read_project_mock.return_value)

    @mock.patch('bigflow.build.operate.build_image')
    @mock.patch('bigflow.build.spec.read_project_spec')
    def test_should_call_build_image_command_through_CLI(
        self,
        read_project_mock: mock.Mock,
        build_image_mock: mock.Mock,
    ):
        # when
        cli(['build-image'])

        # then
        read_project_mock.assert_called_once()
        build_image_mock.assert_any_call(read_project_mock.return_value)

    @mock.patch('bigflow.build.operate.build_dags')
    @mock.patch('bigflow.build.spec.read_project_spec')
    @freezegun.freeze_time("2001-02-03 15:00:00")
    def test_should_call_build_dags_command_through_CLI(
        self,
        read_project_mock: mock.Mock,
        build_dags_mock: mock.Mock,
    ):
        # when
        cli(['build-dags'])

        # then
        read_project_mock.assert_called_once()
        build_dags_mock.assert_any_call(
            read_project_mock.return_value,
            start_time="2001-02-03 15:00:00",
            workflow_id=None,
        )

    @mock.patch('bigflow.cli.get_version')
    def test_should_call_cli_project_version_command(self, get_version):
        # when
        cli(['project-version'])

        # then
        get_version.assert_called_once()

    @mock.patch('bigflow.cli.get_version')
    def test_should_call_cli_project_version_command_by_alias(self, get_version):
        # when
        cli(['pv'])

        # then
        get_version.assert_called_once()

    @mock.patch('bigflow.cli.release')
    def test_should_call_cli_release_command_with_no_args(self, release):
        # when
        cli(['release'])

        # then
        release.assert_called_once_with(None)

    @mock.patch('bigflow.cli.release')
    def test_should_call_cli_release_command_with_identity_file(self, release):
        # when
        cli(['release', '--ssh-identity-file', 'path/to/identity_file'])

        # then
        release.assert_called_once_with('path/to/identity_file')

    @mock.patch('bigflow.cli.release')
    def test_should_call_cli_release_command_with_identity_file_parameter_shortcut(self, release):
        # when
        cli(['release', '-i', 'path/to/identity_file'])

        # then
        release.assert_called_once_with('path/to/identity_file')

    @mock.patch('bigflow.cli.gcloud_project_list')
    @mock.patch('bigflow.cli.project_id_input')
    def test_should_gcp_project_flow_return_project_from_user(self, project_id_input_mock, gcloud_project_list_mock):
        # given
        gcloud_project_list_mock.return_value = '''PROJECT_ID                      NAME                            PROJECT_NUMBER
some-project-id                            SOME PROJECT                   047902537028
another-project-id                         ANOTHER PROJECT                002242200764'''
        project_id_input_mock.return_value = 'some-project-id'
        # when
        project = gcp_project_flow(0)

        # then
        self.assertEqual(project, 'some-project-id')

    @mock.patch('bigflow.cli.get_default_project_from_gcloud')
    @mock.patch('bigflow.cli.gcloud_project_list')
    @mock.patch('bigflow.cli.project_id_input')
    def test_should_gcp_project_flow_return_project_from_gcloud_if_project_not_provided_by_user(self,
                                                                                                project_id_input_mock,
                                                                                                gcloud_project_list_mock,
                                                                                                get_default_project_from_gcloud_mock):
        # given
        gcloud_project_list_mock.return_value = '''PROJECT_ID                      NAME                            PROJECT_NUMBER
some-project-id                            SOME PROJECT                   047902537028
another-project-id                         ANOTHER PROJECT                002242200764'''
        project_id_input_mock.return_value = ''
        get_default_project_from_gcloud_mock.return_value = 'another-project-id'
        # when
        project = gcp_project_flow(0)

        # then
        self.assertEqual(project, 'another-project-id')

    @mock.patch('bigflow.cli.gcloud_project_list')
    @mock.patch('bigflow.cli.project_id_input')
    def test_should_gcp_project_flow_allow_user_to_enter_project_again_if_wrong_project_passed(self,
                                                                                               project_id_input_mock,
                                                                                               gcloud_project_list_mock):
        # given
        gcloud_project_list_mock.return_value = '''PROJECT_ID                      NAME                            PROJECT_NUMBER
some-project-id                            SOME PROJECT                   047902537028
another-project-id                         ANOTHER PROJECT                002242200764'''
        project_id_input_mock.side_effect = ['fake-project', 'some-project-id']
        # when
        project = gcp_project_flow(0)

        # then
        self.assertEqual(project, 'some-project-id')

    @mock.patch('bigflow.cli._cli_start_project')
    def test_should_call_cli_start_project_command(self, cli_start_project_mock):
        # when
        cli(['start-project'])

        # then
        self.assertEqual(cli_start_project_mock.call_count, 1)

    @mock.patch('builtins.print')
    def test_should_call_cli_logs_command(self, print_mock):
        # when
        root_package = TESTS_DIR / "cli_logs_regular_workflows"
        cli_logs(root_package)

        # then
        print_mock.assert_called_with(
            '\n*************************LOGS LINK*************************\nInfrastructure logs:'
            '\nsome-project-id: https://console.cloud.google.com/logs/query;query=%28'
            'severity%3E%3D%22INFO%22%0A'
            'resource.type%3D%22k8s_pod%22%0A%22'
            'Error%3A%22%29%0AOR%0A%28'
            'resource.type%3D%22dataflow_step%22%0A'
            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
            'severity%3E%3D%22INFO%22%0A%29\n'
            'another-project-id: https://console.cloud.google.com/logs/query;query=%28'
            'severity%3E%3D%22INFO%22%0A'
            'resource.type%3D%22k8s_pod%22%0A%22'
            'Error%3A%22%29%0AOR%0A%28'
            'resource.type%3D%22dataflow_step%22%0A'
            'log_name%3D%22projects%2Fanother-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_1%22%0Alabels.workflow_id%3D%22ID_1%22%0A'
            '\nID_2: https://console.cloud.google.com/logs/query;query='
            'logName%3D%22projects%2Fanother-project-id%2Flogs%2FID_2%22%0Alabels.workflow_id%3D%22ID_2%22%0A'
            '\n***********************************************************')

    @mock.patch('builtins.print')
    def test_should_call_cli_logs_and_use_log_name_if_provided(self, print_mock):
        # when
        root_package = TESTS_DIR / "cli_logs_log_name_workflow"
        cli_logs(root_package)

        # then
        print_mock.assert_called_with(
            '\n*************************LOGS LINK*************************\nInfrastructure logs:\nsome-project-id: '
            'https://console.cloud.google.com/logs/query;query=%28'
            'severity%3E%3D%22INFO%22%0A'
            'resource.type%3D%22k8s_pod%22%0A%22'
            'Error%3A%22%29%0AOR%0A%28'
            'resource.type%3D%22dataflow_step%22%0A'
            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
            'logName%3D%22projects%2Fsome-project-id%2Flogs%2Fname-log%22%0A'
            'labels.workflow_id%3D%22ID_1%22%0A\n***********************************************************')

    @mock.patch('builtins.print')
    def test_should_deduplicate_projects_id(self, print_mock):
        # when
        root_package = TESTS_DIR / "cli_logs_duplicated_workflows"
        cli_logs(root_package)

        # then
        print_mock.assert_called_with(
            '\n*************************LOGS LINK*************************\nInfrastructure logs:'
            '\nsome-project-id: https://console.cloud.google.com/logs/query;query=%28'
            'severity%3E%3D%22INFO%22%0A'
            'resource.type%3D%22k8s_pod%22%0A%22'
            'Error%3A%22%29%0AOR%0A%28'
            'resource.type%3D%22dataflow_step%22%0A'
            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_1%22%0A'
            'labels.workflow_id%3D%22ID_1%22%0A\nID_2: https://console.cloud.google.com/logs/query;query='
            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_2%22%0Alabels.workflow_id%3D%22ID_2%22%0A'
            '\n***********************************************************')

    def test_should_raise_exception_if_no_workflow_with_log_config_found(self):
        root_package = TESTS_DIR / "i_do_not_exist_at_all"
        with self.assertRaises(Exception) as e:
            cli_logs(root_package)
            self.assertEqual(str(e.exception), 'Found no workflows with configured logging.')

    def test_should_throw_if_log_module_is_not_installed(self):
        # given
        with mock.patch.dict('sys.modules', {'bigflow.log': None}):
            # when
            with self.assertRaises(Exception):
                cli(['logs'])

    def _expected_default_dags_dir(self):
        return (Path(os.getcwd()) / '.dags').as_posix()

    def _touch_file(self, file_name: str, content: str = '', directory: Optional[str] = None):
        if directory:
            workdir = Path(os.path.join(os.getcwd(), directory))
            workdir.mkdir(exist_ok=True)
        else:
            workdir = Path(os.getcwd())
        f = workdir / file_name
        #self.addCleanup(f.unlink)
        if f.exists():
            # FIXME: Refactor tests - copy workdir into temp directory.
            orig = f.read_bytes()
            self.addCleanup(f.write_bytes, orig)
        else:
            self.addCleanup(f.unlink)
        f.touch()
        f.write_text(content)
        return f

    @mock.patch('bigflow.cli._cli_build')
    @mock.patch.object(logging.root, 'handlers', new=[])
    def test_should_enble_info_logging(self, cli_build):
        # when
        cli(["build"])

        # then
        self.assertEqual(len(logging.root.handlers), 1)
        self.assertIsInstance(logging.root.handlers[0], logging.StreamHandler)
        self.assertEqual(logging.root.level, logging.INFO)

    @mock.patch('bigflow.cli._cli_build')
    @mock.patch.object(logging.root, 'handlers', new=[])
    def test_should_enble_debug_logging_when_verbose_flag_is_specified(self, cli_build):
        # when
        cli(["--verbose", "build"])

        # then
        self.assertEqual(len(logging.root.handlers), 1)
        self.assertIsInstance(logging.root.handlers[0], logging.StreamHandler)
        self.assertEqual(logging.root.level, logging.DEBUG)



from bigflow.commons import public

@public()
def setup(**kwargs):
    """Specify parameters of Bigflow project. Must me called from `setup.py`
    """
    # Laziliy import 'bigflow.dist' as it depends on 'setuptools'
    import bigflow.build.dist as d
    return d.setup(**kwargs)


from bigflow.build.legacy import *


import shutil
import typing

from pathlib import Path
from datetime import datetime

from bigflow import commons
from bigflow.workflow import DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS


def clear_dags_output_dir(workdir: str):
    dags_dir_path = get_dags_output_dir(workdir)

    print("clearing dags_output_dir", str(dags_dir_path.resolve()))
    shutil.rmtree(str(dags_dir_path.resolve()))


def secret_template(secret: str) -> str:
    return f"secret.Secret(deploy_type='env', deploy_target='{secret}', secret='{secret.replace('_', '-')}', key='{secret}')"


def generate_dag_file(workdir: str,
                      docker_repository: str,
                      workflow,
                      start_from: typing.Union[datetime, str],
                      build_ver: str,
                      root_package_name: str) -> str:
    start_from = _str_to_datetime(start_from)

    print(f'start_from: {start_from}')
    print(f'build_ver: {build_ver}')
    print(f'docker_repository: {docker_repository}')

    dag_deployment_id = get_dag_deployment_id(workflow.workflow_id, start_from, build_ver)
    dag_file_path = get_dags_output_dir(workdir) / (dag_deployment_id + '_dag.py')
    start_date_as_str = repr(workflow.start_time_factory(start_from))

    print(f'dag_file_path: {dag_file_path.resolve()}')

    dag_chunks = []

    dag_chunks.append("""
import datetime
from airflow import DAG
from airflow.contrib.operators import kubernetes_pod_operator
from airflow.contrib.kubernetes import secret

default_args = {{
            'owner': 'airflow',
            'depends_on_past': {depends_on_past},
            'start_date': {start_date_as_str},
            'email_on_failure': False,
            'email_on_retry': False,
            'execution_timeout': datetime.timedelta(seconds={execution_timeout_sec}),
}}

dag = DAG(
    '{dag_id}',
    default_args=default_args,
    max_active_runs=1,
    schedule_interval='{schedule_interval}'
)
""".format(dag_id=dag_deployment_id,
           start_date_as_str=start_date_as_str,
           schedule_interval=workflow.schedule_interval,
           depends_on_past=workflow.depends_on_past,
           execution_timeout_sec=DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS))

    def get_job(workflow_job):
        return workflow_job.job

    def build_dag_operator(workflow_job, dependencies):
        job = get_job(workflow_job)
        job_var = "t" + str(job.id)
        task_id = job.id.replace("_", "-")

        execution_timeout_sec = commons.as_timedelta(
            getattr(job, 'execution_timeout_sec', None)
            or DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS)

        dag_chunks.append("""
{job_var} = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='{task_id}',
    name='{task_id}',
    cmds=['bf'],
    arguments=['run', '--job', '{bf_job}', '--runtime', '{{{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}}}', '--project-package', '{root_folder}', '--config', '{{{{var.value.env}}}}'],
    namespace='default',
    image='{docker_image}',
    is_delete_operator_pod=True,
    retries={retries},
    retry_delay=datetime.timedelta(seconds={retry_delay}),
    dag=dag,
    secrets={secrets_definition},
    execution_timeout={execution_timeout_sec!r})
""".format(job_var=job_var,
          task_id=task_id,
          docker_image = commons.build_docker_image_tag(docker_repository, build_ver),
          bf_job= workflow.workflow_id+"."+job.id,
          root_folder=root_package_name,
          retries=job.retry_count if hasattr(job, 'retry_count') else 3,
          retry_delay=job.retry_pause_sec if hasattr(job, 'retry_pause_sec') else 60,
          secrets_definition=f'[{", ".join([secret_template(secret) for secret in workflow.secrets])}]',
          execution_timeout_sec=execution_timeout_sec,
          ))

        for d in dependencies:
            up_job_var = "t" + str(get_job(d).id)
            dag_chunks.append("{job_var}.set_upstream({up_job_var})".format(job_var=job_var, up_job_var=up_job_var))

    workflow._call_on_graph_nodes(build_dag_operator)

    dag_file_content = '\n'.join(dag_chunks) + '\n'
    dag_file_path.write_text(dag_file_content)

    return dag_file_path.as_posix()


def get_dag_deployment_id(workflow_name: str,
                          start_from: str,
                          build_ver: str):
    return '{workflow_name}__v{ver}__{start_from}'.format(
        workflow_name=workflow_name,
        ver=build_ver.translate(str.maketrans(".-+", "___")),
        start_from=_str_to_datetime(start_from).strftime('%Y_%m_%d_%H_%M_%S')
    )


def get_dags_output_dir(workdir: str) -> Path:
    dags_dir_path = Path(workdir) / '.dags'

    if not dags_dir_path.exists():
        dags_dir_path.mkdir()

    return dags_dir_path


def _str_to_datetime(dt: typing.Union[str, datetime]):
    if isinstance(dt, datetime):
        return dt
    elif len(dt) <= 10:
        return datetime.strptime(dt, "%Y-%m-%d")
    else:
        return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")



import logging
import typing as T
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pathlib import Path

import requests
from google.cloud.storage import Bucket
from google.oauth2 import credentials
from google.cloud import storage

import bigflow.commons as bf_commons
from .commons import decode_version_number_from_file_name, remove_docker_image_from_local_registry, build_docker_image_tag


logger = logging.getLogger(__name__)


def load_image_from_tar(image_tar_path: str) -> str:
    logger.info("Load docker image from %s...", image_tar_path)
    for line in bf_commons.run_process(['docker', 'load', '-i', image_tar_path]).split('\n'):
        if 'Loaded image ID:' in line:
            return line.split()[-1].split(':')[-1]
    raise ValueError(f"Can't load image: {image_tar_path}")


def tag_image(image_id: str, repository: str, tag: str) -> str:
    return bf_commons.run_process(["docker", "tag", image_id,  f"{repository}:{tag}"])


def deploy_docker_image(
        image_tar_path: str,
        docker_repository: str,
        auth_method: Literal['local_account', 'vault'] = 'local_account',
        vault_endpoint: T.Optional[str] = None,
        vault_secret: T.Optional[str] = None,
) -> str:
    build_ver = decode_version_number_from_file_name(Path(image_tar_path))
    build_ver = build_ver.replace("+", "-")  # fix local version separator
    image_id = load_image_from_tar(image_tar_path)
    try:
        return _deploy_image_loaded_to_local_registry(build_ver, docker_repository, image_id, auth_method,
                                                      vault_endpoint, vault_secret)
    finally:
        remove_docker_image_from_local_registry(build_docker_image_tag(docker_repository, build_ver))


def _deploy_image_loaded_to_local_registry(
    build_ver: str,
    docker_repository: str,
    image_id: str,
    auth_method: Literal['local_account', 'vault'],
    vault_endpoint: str,
    vault_secret: str,
) -> str:
    tag_image(image_id, docker_repository, build_ver)

    docker_image = docker_repository + ":" + build_ver
    docker_image_latest = docker_repository + ":latest"

    logger.info("Deploying docker image tag=%s auth_method=%s", docker_image, auth_method)

    if auth_method == 'local_account':
        bf_commons.run_process(['gcloud', 'auth', 'configure-docker'])
    elif auth_method == 'vault':
        oauthtoken = get_vault_token(vault_endpoint, vault_secret)
        bf_commons.run_process(
            ['docker', 'login', '-u', 'oauth2accesstoken', '--password-stdin', 'https://eu.gcr.io'],
            input=oauthtoken,
        )
    else:
        raise ValueError('unsupported auth_method: ' + auth_method)

    bf_commons.run_process(['docker', 'push', docker_image])  # TODO(anjensan) use docker_image_latest

    return docker_image


def deploy_dags_folder(
        dags_dir: str,
        dags_bucket: str,
        project_id: str,
        clear_dags_folder: bool = False,
        auth_method: Literal['local_account', 'vault'] = 'local_account',
        vault_endpoint: T.Optional[str] = None,
        vault_secret: T.Optional[str] = None,
        gs_client: T.Optional[storage.Client] = None,
) -> str:
    logger.info("Deploying DAGs folder, auth_method=%s, clear_dags_folder=%s, dags_dir=%s", auth_method, clear_dags_folder, dags_dir)

    client = gs_client or create_storage_client(auth_method, project_id, vault_endpoint, vault_secret)
    bucket = client.bucket(dags_bucket)

    if clear_dags_folder:
        clear_remote_dags_bucket(bucket)

    upload_dags_folder(dags_dir, bucket)
    return dags_bucket


def clear_remote_dags_bucket(bucket: Bucket) -> None:
    i = 0
    for blob in bucket.list_blobs(prefix='dags'):
        if not blob.name in ['dags/', 'dags/airflow_monitoring.py']:
            logger.info("deleting file %s", _blob_uri(blob))
            blob.delete()
            i += 1

    logger.info("%s files deleted", i)


def _blob_uri(blob: storage.Blob) -> str:
    return f"gs://{blob.bucket.name}/{blob.name}"


def upload_dags_folder(dags_dir: str, bucket: Bucket) -> None:
    dags_dir_path = Path(dags_dir)

    def upload_file(local_file_path, target_file_name):
        blob = bucket.blob(target_file_name)
        blob.upload_from_filename(local_file_path, content_type='application/octet-stream')
        logger.info("uploading file %s to %s", local_file_path, _blob_uri(blob))

    files = list(filter(Path.is_file, dags_dir_path.iterdir()))
    for f in files:
        upload_file(f.as_posix(), 'dags/' + f.name)

    logger.info("%s files uploaded", len(files))


def create_storage_client(
        auth_method: str,
        project_id: str,
        vault_endpoint: str,
        vault_secret: str,
) -> storage.Client:
    if auth_method == 'local_account':
        return storage.Client(project=project_id)
    elif auth_method == 'vault':
        oauthtoken = get_vault_token(vault_endpoint, vault_secret)
        return storage.Client(project=project_id, credentials=credentials.Credentials(oauthtoken))
    else:
        raise ValueError('unsupported auth_method: ' + auth_method)


def get_vault_token(vault_endpoint: str, vault_secret: str) -> str:
    if not vault_endpoint:
        raise ValueError('vault_endpoint is required')
    if not vault_secret:
        raise ValueError('vault_secret is required')

    headers = {'X-Vault-Token': vault_secret}
    response = requests.get(vault_endpoint, headers=headers, verify=False)

    if response.status_code != 200:
        logger.info(response.text)
        raise ValueError(
            'Could not get vault token, response code: {}'.format(
                response.status_code))

    logger.info("get oauth token from %s status_code=%s", vault_endpoint, response.status_code)
    return response.json()['data']['token']



import datetime
import unittest
import os
import json

from unittest import mock

import google.cloud.dataproc_v1
import google.cloud.storage

import bigflow
import bigflow.dataproc
import bigflow.resources

from bigflow.workflow import DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS

_someobject_tags = set()

def _some_global_func():
    _someobject_tags.add('global_func')


class _SomeObject:

    def __call__(self):
        _someobject_tags.add('__call__')

    def method(self):
        _someobject_tags.add('method')

    @staticmethod
    def static_method():
        _someobject_tags.add('static_method')

    @classmethod
    def class_method(cls):
        _someobject_tags.add('class_method')


class PySparkJobTest(unittest.TestCase):

    def test_generates_driver_for_callable(self):

        for callable, expected_tag in [
            (_some_global_func, 'global_func'),
            (_SomeObject(), '__call__'),
            (_SomeObject().static_method, 'static_method'),
            (_SomeObject().class_method, 'class_method'),
            (_SomeObject().method, 'method'),
        ]:
            # given
            _someobject_tags.clear()

            # when
            script = bigflow.dataproc.generate_driver_script(callable, {})
            exec(script)

            # then
            # callable was serialized/deserialized/called
            self.assertIn(expected_tag, _someobject_tags)

    @mock.patch.dict('os.environ')
    def test_generated_driver_updates_os_environ(self):
        # when
        script = bigflow.dataproc.generate_driver_script(_SomeObject(), {'some-env': "value"})
        exec(script)

        # then
        self.assertEqual(os.environ.get('some-env'), "value")

    def test_uploads_driver_to_gcs(self):
        # given
        driver_script = "some python code"
        driver_path = 'some-driver.py'

        gsbucket = mock.Mock()
        gsbucket.blob = mock.Mock()

        # when
        bigflow.dataproc._upload_driver_script(driver_script, gsbucket, driver_path)

        # then
        gsbucket.blob.assert_called_with(driver_path)
        gsbucket.blob.return_value.upload_from_string.assert_called_once_with(
            data=driver_script,
            content_type='application/octet-stream',
        )

    def test_should_create_new_cluster_for_job(self):

        # given
        cluster_name = "test-cluster"
        project_id = "some-project"
        region = "us-west1"

        cluster_client = mock.Mock()

        # when
        bigflow.dataproc._create_cluster(
            cluster_client,
            project_id,
            region,
            cluster_name,
            requirements=["pkg1==1", "pkg2==2"],
            worker_machine_type='standart-n1',
            worker_num_instances=4,
        )

        # then
        cluster_client.create_cluster.assert_called_once_with(
            project_id=project_id, region=region, cluster=mock.ANY)

        (_, call_kwargs) = cluster_client.create_cluster.call_args
        cluster_data = call_kwargs['cluster']
        cluster_client.create_cluster.return_value.result.assert_called_once()

        self.assertEqual(cluster_data['cluster_name'], cluster_name)
        self.assertEqual(cluster_data['config']['worker_config']['num_instances'], 4)
        self.assertEqual(cluster_data['config']['worker_config']['machine_type_uri'], 'standart-n1')

        self.assertIn(
            ('PIP_PACKAGES', "pkg1==1 pkg2==2"),
            cluster_data['config']['gce_cluster_config']['metadata'],
        )

    def test_should_send_correct_job_request_to_dataproc(self):

        # given
        cluster_name = "test-cluster"
        project_id = "some-project"
        region = "us-west1"
        bucket_id = "the-bucket"
        driver_path = "some-path/driver.py"
        egg_path = "another-path/file.egg"

        dataproc_job_client = mock.Mock()
        dataproc_job_client.submit_job.return_value.reference.job_id = "the-job"

        # when
        job_id = bigflow.dataproc._submit_single_pyspark_job(
            dataproc_job_client=dataproc_job_client,
            project_id=project_id,
            region=region,
            cluster_name=cluster_name,
            bucket_id=bucket_id,
            driver_path=driver_path,
            egg_path=egg_path,
            jar_file_uris=(),
            properties=(),
        )

        # then
        dataproc_job_client.submit_job.assert_called_once_with(
            project_id=project_id, region=region, job=mock.ANY)

        (_, call_kwargs) = dataproc_job_client.submit_job.call_args
        job = call_kwargs['job']

        self.assertEqual(job['placement']['cluster_name'], cluster_name)
        self.assertEqual(job_id, "the-job")

    @mock.patch('google.cloud.storage.Client')
    @mock.patch('google.cloud.dataproc_v1.ClusterControllerClient')
    @mock.patch('google.cloud.dataproc_v1.JobControllerClient')
    @mock.patch('bigflow.build.reflect.build_egg')
    @mock.patch('bigflow.build.reflect.locate_project_path')
    def test_run_pyspark_job(
        self,
        locate_project_path: mock.Mock,
        build_egg_mock: mock.Mock,
        job_controller_client_cls,
        cluster_controller_client_cls,
        storage_client_cls,
    ):
        # given
        context = bigflow.JobContext.make(
            workflow_id="some_workflow",
            runtime="2020-02-02",
        )

        pyspark_job = bigflow.dataproc.PySparkJob(
            'pyspark_job',
            _some_global_func,
            bucket_id="test-bucket",
            gcp_project_id="test-project",
            gcp_region="us-west1",
        )

        cluster_controller = cluster_controller_client_cls.return_value
        job_controller = job_controller_client_cls.return_value

        submit_job_result = mock.Mock()
        submit_job_result.reference.job_id = "the-job"
        job_controller.submit_job = mock.Mock(return_value=submit_job_result)

        job_state = mock.Mock()
        job_state.status.state = google.cloud.dataproc_v1.JobStatus.State.DONE
        job_state.driver_output_resource_uri = "gs://job-output"
        job_controller.get_job = mock.Mock(return_value=job_state)

        # when
        with self.assertLogs() as logs:
            pyspark_job.execute(context)

        # then
        build_egg_mock.assert_called_once_with(locate_project_path.return_value)

        cluster_controller.create_cluster.assert_called_once()
        cluster_controller.delete_cluster.assert_called_once()

        job_controller.submit_job.assert_called_once_with(
            project_id=pyspark_job.gcp_project_id,
            region=pyspark_job.gcp_region,
            job=mock.ANY,
        )

        job = job_controller.submit_job.call_args[1]['job']
        self.assertEqual(job['pyspark_job']['properties'], pyspark_job._prepare_pyspark_properties(context))

        self.assertTrue(any("JOB OUTPUT:" in line for line in logs.output))

    def test_should_pass_job_context_to_callable(self):
        # given
        runtime = datetime.datetime(2020, 1, 2, 3, 4)

        workflow = mock.Mock()
        workflow.workflow_id = "some_workflow"
        workflow.log_config = None

        jc = bigflow.JobContext.make(
            workflow=workflow,
            runtime=runtime,
        )

        callback = mock.Mock()
        job = bigflow.dataproc.PySparkJob(
            "some_job",
            callback,
            bucket_id="no-bucket",
            gcp_project_id="no-project",
            gcp_region="no-region",
            project_name="test",
        )

        # when
        driver = job._prepare_driver_callable(jc)
        driver()

        # then
        callback.assert_called_once()
        (jc2,), _kwargs = callback.call_args

        self.assertIsNone(jc2.workflow)
        self.assertEqual(jc2.workflow_id, workflow.workflow_id)
        self.assertEqual(jc2.runtime, runtime)

    @mock.patch('os.environ')
    @mock.patch('bigflow.log.maybe_init_logging_from_env')
    def test_should_initialize_logging_on_driver(self, maybe_init_logging, os_environ):

        # given
        os.environ = {}
        workflow = mock.Mock()
        workflow.workflow_id = "some_workflow"
        workflow.log_config = {"level": "DEBUG"}

        jc = bigflow.JobContext.make(workflow=workflow)

        callback = id
        job = bigflow.dataproc.PySparkJob(
            "some_job",
            callback,
            bucket_id="no-bucket",
            gcp_project_id="no-project",
            gcp_region="no-region",
            project_name="test",
        )

        # when
        driver_script = job._prepare_driver_script(jc)
        exec(driver_script)

        # then
        self.assertIn('bf_log_config', os.environ)
        self.assertDictEqual(json.loads(os.environ['bf_log_config']), workflow.log_config)
        self.assertEqual(os.environ['bf_workflow_id'], workflow.workflow_id)
        maybe_init_logging.assert_called_once()

    def test_should_initialize_logging_on_workers(self):

        # given
        workflow = mock.Mock()
        workflow.workflow_id = "some_workflow"
        workflow.log_config = {"level": "DEBUG"}

        jc = bigflow.JobContext.make(workflow=workflow)

        callback = lambda context: None
        job = bigflow.dataproc.PySparkJob(
            "some_job",
            callback,
            bucket_id="no-bucket",
            gcp_project_id="no-project",
            gcp_region="no-region",
            project_name="test",
        )

        # when
        props = job._prepare_pyspark_properties(jc)

        # then
        self.assertIn('spark.executorEnv.bf_log_config', props)
        self.assertIn('spark.executorEnv.bf_workflow_id', props)

    def test_should_initialize_job_with_default_execution_timeout(self):
        # when
        job = bigflow.dataproc.PySparkJob(
            "some_job",
            id,
            bucket_id="no-bucket",
            gcp_project_id="no-project",
            gcp_region="no-region",
            project_name="test",
        )

        # then
        self.assertEqual(job.execution_timeout_sec, DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS)

    def test_should_initialize_job_with_execution_timeout(self):
        # when
        job = bigflow.dataproc.PySparkJob(
            "some_job",
            id,
            bucket_id="no-bucket",
            gcp_project_id="no-project",
            gcp_region="no-region",
            execution_timeout_sec=1,
            project_name="test",
        )

        # then
        self.assertEqual(job.execution_timeout_sec, 1)



import os

from unittest import TestCase

from bigflow.configuration import Config, DeploymentConfig
from bigflow.bigquery.dataset_configuration import DatasetConfig


def _set_os_env(value=None, env_var_name='bf_env'):
    for key in ['env', 'bf_env', 'my_namespace_']:
        if key in os.environ:
            del os.environ[key]

    if value:
        os.environ[env_var_name] = value


class TestConfig(TestCase):

    def setUp(self):
        self.old_os_environ = os.environ.copy()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self.old_os_environ)

    def test_should_resolve_explicit_properties_in_simple_config(self):
        # when
        config = Config('dev', {'a':1, 'b':'2'})

        #expect
        self.assertEqual(config.resolve('dev'), {'a':1, 'b':'2', 'env': 'dev'})

    def test_should_resolve_to_os_env_variable_when_property_value_is_None(self):
        # when
        os.environ['bf_b'] = 'x'
        config = Config('dev', {'a': 1, 'b': None})

        # then
        self.assertEqual(config.resolve('dev'), {'a':1, 'b': 'x', 'env': 'dev'})


    def test_should_smartly_resolve_properties_with_placeholders(self):
        # when
        config = Config('dev', {
                                   'project_id': 'dev_project',
                                   'topic': 'topic/{project_id}/my_topic}'
                               }) \
            .add_configuration('prod',
                               {
                                   'project_id': 'prod_project',
                               })
        # then
        self.assertEqual(config.resolve_property('topic', 'dev'),  'topic/dev_project/my_topic}')
        self.assertEqual(config.resolve_property('topic', 'prod'), 'topic/prod_project/my_topic}')

    def test_should_raise_Error_when_config_name_not_found(self):
        with self.assertRaises(ValueError):
            Config('dev', {}).resolve('prod')

    def test_should_raise_Error_when_default_config_requested_but_not_exists(self):
        # when
        _set_os_env('prod')

        with self.assertRaises(ValueError):
            Config('dev', {}, is_master=False).resolve()

    def test_should_resolve_config_given_by_os_env_variable(self):
        # when
        _set_os_env('dev')

        config = Config('prod', {'bb': 1}).add_configuration('dev', {'bb': 2})

        # then
        self.assertEqual(config.resolve(), {'bb': 2, 'env': 'dev'})

    def test_should_give_priority_to_explicit_properties_rather_than_os_env_variables(self):
        # when
        os.environ['bf_bb'] = 'x'
        config = Config('dev', {'bb': 1})

        # then
        self.assertEqual(config.resolve('dev'), {'bb': 1, 'env': 'dev'})
        self.assertEqual(config.resolve(), {'bb': 1, 'env': 'dev'})
        self.assertEqual(config.resolve_property('bb', 'dev'), 1)
        self.assertEqual(config.resolve_property('bb'), 1)

    def test_should_raise_Error_when_required_os_env_variable_is_not_found(self):
        with self.assertRaises(ValueError):
            Config('dev', {'no_such_variable': None}).resolve('dev')

    def test_should_support_hierarchical_configuration(self):
        # when
        config = Config('dev', {'a': 1, 'b': 2})\
            .add_configuration('prod', {'a': 5, 'b':10, 'c':20})

        # then
        self.assertEqual(config.resolve('dev'), {'a': 1, 'b': 2, 'env': 'dev'})
        self.assertEqual(config.resolve('prod'), {'a': 5, 'b': 10, 'c': 20, 'env': 'prod'})

    def test_should_resolve_from_master_config_when_property_is_missing(self):
        # when
        config = Config('dev', {'a': 1})\
            .add_configuration('prod', {})

        # expect
        self.assertEqual(config.resolve('dev'),  {'a': 1, 'env': 'dev'})
        self.assertEqual(config.resolve('prod'), {'a': 1, 'env': 'prod'})

    def test_should_use_bg_as_the_default_environment_variables_prefix(self):
        _set_os_env('prod', 'bf_env')
        # when
        config = Config('dev', {'a': 'dev1'})\
            .add_configuration('prod', {'a': 'prod2'})

        #expect
        self.assertEqual(config.resolve(), {'a': 'prod2', 'env': 'prod'})

    def test_should_raise_Error_no_explicit_env_is_given_nor_default_env_is_defined(self):
        with self.assertRaises(ValueError):
            Config('dev', {'a': 1}, is_default=False).resolve()

    def test_should_raise_Error_when_more_than_one_default_env_is_defined(self):
        with self.assertRaises(ValueError):
            Config('dev', {'a': 1}, is_default=True)\
                .add_configuration('prod', {'a': 2}, is_default=True)


    def test_should_use_default_env_from_master_config_when_no_env_is_given(self):
        # when
        config = Config('dev', {'a': 'dev1'})\
            .add_configuration('prod', {'a': 'prod2'})\
            .add_configuration('test', {'a': 'test3'})

        # expect
        self.assertEqual(config.resolve(), {'a': 'dev1', 'env': 'dev'})
        self.assertEqual(config.resolve('dev'), {'a': 'dev1', 'env': 'dev'})
        self.assertEqual(config.resolve('prod'), {'a': 'prod2', 'env': 'prod'})
        self.assertEqual(config.resolve('test'), {'a': 'test3', 'env': 'test'})

        # when
        _set_os_env('prod')
        self.assertEqual(config.resolve(), {'a': 'prod2', 'env': 'prod'})

        _set_os_env('dev')
        self.assertEqual(config.resolve(), {'a': 'dev1', 'env': 'dev'})

        _set_os_env('test')
        self.assertEqual(config.resolve(), {'a': 'test3', 'env': 'test'})

    def test_should_use_default_env_from_secondary_config_when_no_env_is_given(self):
        # when
        config = Config('dev', {'a': 'dev1'}, is_default=False)\
            .add_configuration('prod', {'a': 'prod2'}, is_default=True)\
            .add_configuration('test', {'a': 'test3'})

        # expect
        self.assertEqual(config.resolve(), {'a': 'prod2', 'env': 'prod'})
        self.assertEqual(config.resolve('dev'), {'a': 'dev1', 'env': 'dev'})
        self.assertEqual(config.resolve('prod'), {'a': 'prod2', 'env': 'prod'})
        self.assertEqual(config.resolve('test'), {'a': 'test3', 'env': 'test'})

    def test_should_use_default_env_from_master_config_in_DatasetConfig(self):
        # when
        config = DatasetConfig('dev',
                               project_id='my_dev_project_id',
                               properties={
                                   'a': 'dev_',
                               }) \
            .add_configuration('prod',
                               project_id='my_prod_project_id',
                               properties={
                                   'a': 'prod_',
                               })

        # then
        self.assertEqual(config.resolve_project_id(), 'my_dev_project_id')
        self.assertEqual(config.resolve_extra_properties(), { 'a': 'dev_'})

    def test_should_use_default_env_from_secondary_config_in_DatasetConfig(self):
        # when
        config = DatasetConfig('dev',
                               project_id='my_dev_project_id',
                               properties={
                                   'a': 'dev_',
                               },
                               is_default=False) \
            .add_configuration('prod',
                               project_id='my_prod_project_id',
                               properties={
                                   'a': 'prod_',
                               },
                               is_default=True)

        # then
        self.assertEqual(config.resolve_project_id(), 'my_prod_project_id')
        self.assertEqual(config.resolve_extra_properties(), {'a': 'prod_'})

    def test_should_model_certain_properties_in_DatasetConfig(self):
        # when
        config = DatasetConfig('dev',
                               project_id='my_dev_project_id',
                               dataset_name='None',
                               internal_tables=['table_1', 'table_2'],
                               properties={
                                   'page-routes-v2': 'page-routes-v2-{env}',
                                   'datastore_project_id': 'my_dev_datastore_project',
                                   'output_topic_name': 'projects/{project_id}/topics/mini',
                                   'window_period_seconds': '60'
                               }) \
            .add_configuration('prod',
                               project_id='my_prod_project_id',
                               external_tables={'table': 'table_1'},
                               properties={
                                   'datastore_project_id': 'my_prod_datastore_project'
                               })

        # then
        self.assertEqual(config.resolve_project_id(),            'my_dev_project_id')
        self.assertEqual(config.resolve_project_id('dev'),       'my_dev_project_id')
        self.assertEqual(config.resolve_dataset_name('dev'),     'None')
        self.assertEqual(config.resolve_internal_tables('dev'),  ['table_1', 'table_2'])
        self.assertEqual(config.resolve_external_tables('dev'),  {})
        self.assertEqual(config.resolve_extra_properties('dev'), {
                                           'page-routes-v2': 'page-routes-v2-dev',
                                           'datastore_project_id': 'my_dev_datastore_project',
                                           'output_topic_name': 'projects/my_dev_project_id/topics/mini',
                                           'window_period_seconds': '60'
                                       })

        self.assertEqual(config.resolve_project_id('prod'),      'my_prod_project_id')
        self.assertEqual(config.resolve_dataset_name('prod'),    'None')
        self.assertEqual(config.resolve_internal_tables('prod'), ['table_1', 'table_2'])
        self.assertEqual(config.resolve_external_tables('prod'), {'table': 'table_1'})
        self.assertEqual(config.resolve_extra_properties('prod'), {
                                           'page-routes-v2': 'page-routes-v2-prod',
                                           'datastore_project_id': 'my_prod_datastore_project',
                                           'output_topic_name': 'projects/my_prod_project_id/topics/mini',
                                           'window_period_seconds': '60'
                                       })

    def test_should_resolve_env_variables_via_resolve_method(self):
        # when
        config = Config('dev', {'a': 'dev1'})

        #then
        self.assertEqual(config.resolve(), {'a': 'dev1', 'env': 'dev'})
        self.assertEqual(config.resolve('dev'), {'a': 'dev1', 'env': 'dev'})

        # when
        os.environ['bf_b'] = 'b_from_env'

        # then
        self.assertEqual(config.resolve(), {'a': 'dev1', 'b': 'b_from_env', 'env': 'dev'})
        self.assertEqual(config.resolve('dev'), {'a': 'dev1', 'b': 'b_from_env', 'env': 'dev'})


class DeploymentConfigTestCase(TestCase):
    def test_should_use_os_environment_variable_prefix_if_given(self):
        # when
        os.environ['my_namespace_b'] = 'x'
        config = DeploymentConfig('dev', {'b': None}, environment_variables_prefix='my_namespace_')

        # then
        self.assertEqual(config.resolve('dev'), {'b': 'x', 'env': 'dev'})

    def test_should_resolve_config_given_by_os_env_variable_with_prefix(self):
        # when
        _set_os_env('test', 'my_namespace_env')

        config = DeploymentConfig('prod', {'bb': 'prod'}, environment_variables_prefix='my_namespace_') \
            .add_configuration('test', {'bb': 'test'})

        # then
        self.assertEqual(config.resolve(), {'bb': 'test', 'env': 'test'})



"""This file contains legacy shims to enable building of old (v1.0) bigflow projects"""

import setuptools
from pathlib import Path

from bigflow.commons import public
from bigflow.build.spec import read_project_spec_nosetuppy
from bigflow.build.dist import projectspec_to_setuppy_kwargs, _maybe_dump_setup_params


_reason = "Use `bigflow.build.setup` instead"


@public(deprecate_reason=_reason)
def auto_configuration(project_name: str, project_dir: Path = Path('.').parent, **kwargs) -> dict:
    project_dir = project_dir or Path.cwd()
    return {
        'project_name': project_name,
        'project_dir': project_dir,
        **kwargs,
    }


@public(deprecate_reason=_reason)
def project_setup(project_name, project_dir=None, **kwargs) -> dict:
    _maybe_dump_setup_params({'name': project_name, **kwargs})
    prj = read_project_spec_nosetuppy(project_dir=project_dir, name=project_name, **kwargs)
    return projectspec_to_setuppy_kwargs(prj)


@public(deprecate_reason=_reason)
def default_project_setup(project_name: str, project_dir=None, **kwargs):
    setuptools.setup(**project_setup(**auto_configuration(project_name, project_dir, **kwargs)))



import logging
import unittest
import textwrap

from typing import Any, Dict, List
from test import mixins

from unittest import mock

from bigflow.build import spec


class ReadSpecTestCase(
    mixins.TempCwdMixin,
    unittest.TestCase,
):

    def test_should_raise_error_when_cant_find_deployment_config(self):
        # then
        with self.assertRaises(ValueError):
            # when
            spec.get_docker_repository_from_deployment_config(self.cwd / 'unknown.py')

    @mock.patch('bigflow.version.get_version')
    def test_fallback_to_detect_version(self, get_version_mock: mock.Mock):
        # given
        get_version_mock.side_effect = ValueError()

        # when
        with self.assertLogs(level=logging.WARNING) as logs:
            v = spec.secure_get_version()

        # then
        self.assertTrue(len(logs.records))
        self.assertEqual(v, "INVALID")

    def test_docker_repository_not_in_lower_case(self):

        # given
        dc = self.cwd / "deployment_config.py"
        dc.write_text(textwrap.dedent("""
            import bigflow
            deployment_config = bigflow.Config(name='dev', properties={'docker_repository': "Docker_Repository"})
        """))

        # then
        with self.assertRaises(ValueError):
            # when
            spec.get_docker_repository_from_deployment_config(dc)



class _BaseRealProjectTest(
    mixins.SubprocessMixin,
    mixins.PrototypedDirMixin,
    mixins.BigflowInPythonPathMixin,
    mixins.ABCTestCase,
    unittest.TestCase,
):
    __test__ = False

    expected_name: str
    expected_packages: List[str]
    expected_metainfo: Dict[str, Any]

    expected_requires: List[str] = [
        "datetime_truncate==1.1.0",
        "six==1.15.0",
        "typing==3.7.4.3",
    ]

    def test_read_spec_from_setuppy(self):

        # when
        s = spec.read_project_spec(self.cwd)

        # then
        self.assertEqual(s.project_dir, self.cwd)
        self.assertEqual(s.name, self.expected_name)
        self.assertEqual(s.metainfo, self.expected_metainfo)

        self.assertEqual(s.docker_repository, "test_repository")
        self.assertCountEqual(s.requries, self.expected_requires)
        self.assertCountEqual(
            s.packages,
            self.expected_packages,
        )

    def test_fill_defaults(self):

        # given
        s = spec.read_project_spec(self.cwd)
        s.name = "new-name"
        s.version = "3.4.5+local"
        # s.data_files = [("one", ["two", "three/file"])]  # https://github.com/uiri/toml/issues/270
        s.deployment_config_file = "no_deployment.py"
        s.project_requirements_file = "req.txt"
        s.resources_dir = "no_resources"
        s.requries = ["my-lib"]
        s.packages = ["a", "b"]
        s.metainfo = {
            'author': "Me",
            'author_email': "no@example.org",
            'description': "No",
            'url': "http://example.org/myproject",
        }

        # when
        spec.add_spec_to_pyproject_toml(self.cwd / "pyproject.toml", s)

        # then
        ss = spec.read_project_spec_nosetuppy(self.cwd)
        for f in [
            'name',
            'version',
            'deployment_config_file',
            'project_requirements_file',
            'resources_dir',
            'requries',
            'packages',
            'metainfo',
            'data_files',
        ]:
            self.assertEqual(getattr(s, f), getattr(ss, f), f"field {f} should be same")


class SpecBigflowV10TestCase(_BaseRealProjectTest):
    __test__ = True
    proto_dir = "bf-projects/bf_simple_v10"

    expected_name = "bf-simple-v10"
    expected_packages = ["simple_v10"]
    expected_metainfo = {}


class SpecBigflowV11TestCase(_BaseRealProjectTest):
    __test__ = True
    proto_dir = "bf-projects/bf_simple_v11"

    expected_name = "bf-simple-v11"
    expected_packages = ["simple_v11"]
    expected_metainfo = dict(
        author="Bigflow UnitTest",
        description="Sample bigflow project",
        url="http://example.org",
    )


class SpecBigflowV12TestCase(_BaseRealProjectTest):
    __test__ = True
    proto_dir = "bf-projects/bf_simple_v12"

    expected_name = "bf-simple-v12"
    expected_packages = ["simple_v12_one", "simple_v12_two"]
    expected_metainfo = dict(
        author="Bigflow UnitTest",
        description="Sample bigflow project",
        url="http://example.org",
    )





from bigflow.scaffold import templating
import os
import tempfile

from unittest import TestCase, mock
from pathlib import Path
from bigflow.resources import *


class FindAllResourcesTestCase(TestCase):
    def test_should_find_all_resource_file_paths(self):
        with tempfile.TemporaryDirectory() as d:
            dp = Path(d)
            for f in [
                dp / "file1",
                dp / "file2",
                dp / "file.txt",
                dp / "dir" / "file3",
                dp / "dir" / "subdir" / "file4",
            ]:
                f.parent.mkdir(parents=True, exist_ok=True)
                f.touch()

            resources = list(find_all_resources(dp))

            # then
            self.assertSequenceEqual(sorted(resources), sorted([
                f"{dp.name}/file1",
                f"{dp.name}/file2",
                f"{dp.name}/file.txt",
                f"{dp.name}/dir/file3",
                f"{dp.name}/dir/subdir/file4",
            ]))


class FindFileTestCase(TestCase):
    def test_should_find_file_going_up_through_hierarchy(self):
        # when
        result_path = find_file('LICENSE', Path(__file__))

        # then
        self.assertEqual(result_path, Path(__file__).parent.parent / 'LICENSE')

    def test_should_raise_error_when_file_not_found(self):
        # then
        with self.assertRaises(ValueError) as e:
            # when
            find_file('LICENSE', Path(__file__), max_depth=1)


class GetResourceAbsolutePathTestCase(TestCase):
    def test_should_find_resource_path(self):
        # when
        result_path = get_resource_absolute_path('test_resource', Path(__file__))

        # then
        self.assertEqual(result_path, Path(__file__).parent / 'resources' / 'test_resource')

        # when
        result_path = get_resource_absolute_path('test_resource')

        # then
        self.assertEqual(result_path, Path(__file__).parent / 'resources' / 'test_resource')

    def test_should_raise_error_when_resource_not_found(self):
        # then
        with self.assertRaises(ValueError) as e:
            # when
            get_resource_absolute_path('unknown_resource', Path(__file__))


class FindSetupTestCase(TestCase):
    def test_should_find_setup(self):
        # when
        setup_path = find_setup(Path(__file__))

        # then
        self.assertEqual(setup_path, Path(__file__).parent.parent / 'setup.py')

    @mock.patch('bigflow.resources.find_file')
    def test_should_retry_n_times(self, find_file_mock):
        # given
        find_file_mock.side_effect = self.raise_value_error

        # then
        with self.assertRaises(ValueError) as e:
            # when
            find_setup(Path(__file__), retries_left=2, sleep_time=0.01)

        # and
        self.assertEqual(find_file_mock.call_count, 3)

    def raise_value_error(self, *args, **kwargs):
        raise ValueError('Setup file not found')


class CreateFileIfNotExistsTestCase(TestCase):
    def setUp(self):
        self.example_file = Path(__file__).parent / 'resources' / 'example_file'
        if self.example_file.exists():
              os.remove(self.example_file)

    def test_should_create_existent_or_non_existent_file(self):
        # when
        create_file_if_not_exists(self.example_file, 'test file')

        # then
        self.assertTrue(self.example_file.read_text(), 'test file')

        # when trying to create existing
        create_file_if_not_exists(self.example_file, 'test file')

        # then
        self.assertTrue(self.example_file.read_text(), 'test file')


class CreateSetupBodyTestCase(TestCase):
    def test_should_create_setup_body(self):
        # when
        body = create_setup_body('example_project')

        # then
        self.assertEqual(body, f'''
import setuptools

setuptools.setup(
        name='example_project',
        version='0.1.0',
        packages=setuptools.find_packages(
                exclude=tuple(
                        p for p in setuptools.find_packages()
                        if not p.startswith('example_project'))))
''')




import logging
import unittest

from pathlib import Path

from test import mixins

import bigflow.build.pip as bf_pip


class PipToolsTestCase(
    mixins.TempCwdMixin,
    mixins.FileUtilsMixin,
    unittest.TestCase,
):

    def test_should_compile_requirements(self):
        # given
        req_in = self.cwd / "req.in"
        req_txt = self.cwd / "req.txt"
        req_in.write_text("pandas>=1.1")

        # when
        bf_pip.pip_compile(req_in)

        # then
        reqs = req_txt.read_text()
        self.assertIn("pandas==", reqs)
        self.assertIn("", reqs)

    def test_should_ignore_pip_flags(self):
        # given
        req_in = self.cwd / "req.in"
        req_txt = self.cwd / "req.txt"

        req_in.write_text("""
        --extra-index-url https://example.org/python/repo
        pandas
        """)

        # when
        bf_pip.pip_compile(req_in)

        # then
        reqs = req_txt.read_text()
        self.assertIn("pandas==", reqs)
        self.assertIn("", reqs)

    def test_detect_when_requirements_was_changed(self):

        # given
        req_in = self.cwd / "req.in"
        req_in.write_text("pandas>=1.1")

        # when
        bf_pip.pip_compile(req_in)

        # then
        self.assertFalse(bf_pip.check_requirements_needs_recompile(req_in))

        # when
        req_in.write_text("pandas>=1.1.1,<2")

        # then
        with self.assertLogs(bf_pip.logger, level=logging.WARNING):
            self.assertTrue(bf_pip.check_requirements_needs_recompile(req_in))

    def test_detect_when_requirements_was_changed_included(self):

        # given
        req_in = self.cwd / "req.in"
        sub_in = self.cwd / "sub.in"
        req_in.write_text("-r sub.in # comment")
        sub_in.write_text("pandas>=1.1")

        # when
        bf_pip.pip_compile(req_in)

        # then
        self.assertFalse(bf_pip.check_requirements_needs_recompile(req_in))

        # when
        sub_in.write_text("pandas>=1.1.1,<2")

        # then
        with self.assertLogs(bf_pip.logger, level=logging.WARNING):
            self.assertTrue(bf_pip.check_requirements_needs_recompile(req_in))

    def test_automatically_recompile_requirements(self):
        # given
        req_in = self.cwd / "req.in"
        req_txt = self.cwd / "req.txt"
        req_in.write_text("numpy")

        # when
        bf_pip.maybe_recompile_requirements_file(req_txt)

        # then
        self.assertTrue(req_txt.exists())

    def test_read_all_requirements_from_the_hierarchy(self):
        # given
        (self.cwd / "requirements.txt").write_text("""
            # comments are allowed
            -r requirements_base.txt  # comment
            # a few empty lines
            datetime_truncate==1.1.0  # another # comment with ### inside
        """)

        (self.cwd / "requirements_base.txt").write_text("""
            freezegun==0.3.14
            schedule
        """)

        # when
        requirements = bf_pip.read_requirements(self.cwd / "requirements.txt")

        # then
        self.assertEqual(requirements, [
            'freezegun==0.3.14',
            'schedule',
            'datetime_truncate==1.1.0',
        ])

    def test_generate_pinfile(self):

        # given
        req_in = self.cwd / "req.in"
        req_txt = self.cwd / "req.txt"
        pin_in = self.cwd / "pin.in"

        req_in.write_text("""
            requests==2.25.1
        """)

        bf_pip.pip_compile(req_in)

        # when
        with self.assertLogs(level=logging.WARNING):
            bf_pip.generate_pinfile(
                req_in,
                pin_in,
                lambda: ["idna==2.0", "chardet==4", "urllib3~=1.26,<1.26.3"],
            )

        # then
        self.assertFileContentRegex(pin_in, r"(?m)^## idna==2.0 +")
        self.assertFileContentRegex(pin_in, r"(?m)^urllib3~=1.26,<1.26.3$")
        self.assertFileContentRegex(pin_in, r"(?m)^chardet==4$")

        self.assertFileContentRegex(req_txt, r"(?m)^requests==2.25.1$")
        self.assertFileContentRegex(req_txt, r"(?m)^idna==2.10$")
        self.assertFileContentRegex(req_txt, r"(?m)^urllib3==1.26.2$")
        self.assertFileContentRegex(req_txt, r"(?m)^chardet==4.0.0$")


#!/usr/bin/env python

import setuptools
import os


with open("README.md", "r") as fh:
    long_description = fh.read()


def read_requirements(name):
    with open(os.path.join("requirements", name)) as f:
        return list(map(str.strip, f))


with open(os.path.join('bigflow', '_version.py'), 'r') as version_file:
    version_globals = {}
    exec(version_file.read(), version_globals)
    version = version_globals['__version__']


setuptools.setup(
    name="bigflow",
    version=version,
    author=u"Chi",
    author_email="chibox-team@allegrogroup.com",
    description="BigQuery client wrapper with clean API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/allegro/bigflow",
    packages=setuptools.find_packages(exclude=('test', 'e2e')),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=read_requirements("base.txt"),
    extras_require={
        'bigquery': read_requirements("bigquery_extras.txt"),
        'log': read_requirements("log_extras.txt"),
        'dataproc': read_requirements("dataproc_extras.txt"),
        'dataflow': read_requirements("dataflow_extras.txt"),
    },
    scripts=["scripts/bf", "scripts/bigflow"],
)



from unittest import TestCase
import os

from bigflow.commons import *


class CliTestCase(TestCase):

    def test_should_decode_version_number_from_file_name(self):
        # given
        f_release = self._touch_file('image-0.1.0.tar')
        f_dev = self._touch_file('image-0.3.0.dev-4b45b638.tar')

        # expect
        self.assertEqual(decode_version_number_from_file_name(f_release), '0.1.0')
        self.assertEqual(decode_version_number_from_file_name(f_dev), '0.3.0.dev-4b45b638')

        f_release.unlink()
        f_dev.unlink()

    def test_should_raise_error_when_given_image_file_is_not_tar(self):
        # given
        f = self._touch_file('image-0.1.0.zip')

        with self.assertRaises(ValueError):
            # when
            decode_version_number_from_file_name(f)

        f.unlink()

    def test_should_raise_error_when_given_file_name_has_wrong_pattern(self):
        # given
        f = self._touch_file('image.0.1.0.tar')

        with self.assertRaises(ValueError):
            # when
            decode_version_number_from_file_name(f)

        f.unlink()

    def test_should_raise_error_when_given_file_does_not_exists(self):
        with self.assertRaises(ValueError):
            # when
            decode_version_number_from_file_name(Path('/Users/image-0.1122123.0.tar'))

    def _touch_file(self, file_name: str, content: str = ''):
        workdir = Path(os.getcwd())
        f = workdir / file_name
        f.touch()
        f.write_text(content)
        return f


import unittest

from test import mixins

import bigflow.build.dev


class ReadProjectParametersLegacyTestCase(
    mixins.PrototypedDirMixin,
    mixins.BigflowInPythonPathMixin,
    unittest.TestCase,
):
    proto_dir = "bf-projects/bf_simple_v10"

    def test_should_read_legacy_project_params(self):
        # when
        params = bigflow.build.dev.read_setuppy_args()

        # then
        self.assertIn('name', params)
        self.assertEqual("bf_simple_v10", params['name'])



class ReadProjectParametersTestCase(
    mixins.PrototypedDirMixin,
    mixins.TempCwdMixin,
    mixins.BigflowInPythonPathMixin,
    unittest.TestCase,
):
    proto_dir = "bf-projects/bf_simple_v11"

    def test_should_read_project_params(self):
        # when
        params = bigflow.build.dev.read_setuppy_args()

        # then
        self.assertEqual(dict(
            name="bf_simple_v11",
            author="Bigflow UnitTest",
            description="Sample bigflow project",
            url="http://example.org",
            version="1.2.3",
        ), params)

    def test_should_read_project_params_from_subdir(self):
        # given
        self.chdir(self.cwd / "sudir" / "subsubdir")

        # when
        params = bigflow.build.dev.read_setuppy_args()

        # then
        self.assertDictContainsSubset({'name': "bf_simple_v11"}, params)

    def test_should_read_project_params_with_module_setup(self):
        # given
        self.chdir(self.cwd / "submodule")
        (self.cwd / "setup.py").write_text("pass")
        (self.cwd / "__init__.py").touch()

        # when
        params = bigflow.build.dev.read_setuppy_args()

        # then
        self.assertDictContainsSubset({'name': "bf_simple_v11"}, params)


"""Read and parse bigflow project configuration (setup.py / pyproject.toml)"""

import functools
import textwrap
import typing
import logging
import typing
import dataclasses
import toml

from pathlib import Path

import setuptools

import bigflow.resources
import bigflow.version

import bigflow.build.pip
import bigflow.build.dev
import bigflow.build.dataflow.dependency_checker

import bigflow.commons as bf_commons


logger = logging.getLogger(__name__)


_ListStr = typing.List[str]

# copied from `distutils.dist.DistributionMetadata`
_PROJECT_METAINFO_KEYS = (
    "name", "version", "author", "author_email",
    "maintainer", "maintainer_email", "url",
    "license", "description", "long_description",
    "keywords", "platforms", "fullname", "contact",
    "contact_email", "classifiers", "download_url",
    # PEP 314
    "provides", "requires", "obsoletes",
)


@dataclasses.dataclass
class BigflowProjectSpec:
    """Parameters of bigflow project."""

    # Path to directory containing 'pyproject.toml' or 'setup.py', transient
    project_dir: Path

    # Basic setuptools info - required or semi-required
    name: str
    version: str
    packages: _ListStr
    requries: _ListStr

    # Setuptools-speicfic but _patched_ (not replaced) by bigflow
    data_files: typing.List

    # Bigflow-specific information and options
    docker_repository: str
    resources_dir: str
    deployment_config_file: str
    project_requirements_file: str

    # Known package metainformation (author, url, description etc, see _PROJECT_METAINFO_KEYS)
    metainfo: typing.Dict[str, typing.Any]

    # Just bypass any unknown options to 'distutils'
    setuptools: typing.Dict[str, typing.Any]


def parse_project_spec(
    project_dir,
    *,
    name,
    docker_repository=None,
    version=None,
    packages=None,
    data_files=None,
    requries=None,
    deployment_config_file="deployment_config.py",
    project_requirements_file="resources/requirements.txt",
    resources_dir="resources",
    **kwargs,

) -> BigflowProjectSpec:
    """Creates instance of BigflowProjectSpec. Populate defaults, coerce values"""

    logger.info("Prepare bigflow project spec...")
    name = name.replace("_", "-")  # PEP8 compliant package names

    docker_repository = docker_repository or get_docker_repository_from_deployment_config(project_dir / deployment_config_file)
    version = version or secure_get_version()
    packages = packages if packages is not None else discover_project_packages(project_dir)
    requries = requries if requries is not None else read_project_requirements(project_dir / project_requirements_file)
    metainfo = {k: kwargs.pop(k) for k in _PROJECT_METAINFO_KEYS if k in kwargs}

    setuptools = kwargs  # all unknown arguments
    if setuptools:
        logger.info("Found unrecognized build parameters: %s", setuptools)

    logger.info("Bigflow project spec is ready")

    return BigflowProjectSpec(
        name=name,
        project_dir=project_dir,
        docker_repository=docker_repository,
        version=version,
        packages=packages,
        requries=requries,
        data_files=data_files,
        resources_dir=resources_dir,
        project_requirements_file=project_requirements_file,
        deployment_config_file=deployment_config_file,
        metainfo=metainfo,
        setuptools=kwargs,
    )


def render_project_spec(prj: BigflowProjectSpec) -> dict:
    """Convertes project spec into embeddable dict"""
    return {
        'name': prj.name,
        'version': prj.version,
        'packages': prj.packages,
        'requries': prj.requries,
        'docker_repository': prj.docker_repository,
        'deployment_config_file': prj.deployment_config_file,
        'project_requirements_file': prj.project_requirements_file,
        # 'data_files': prj.data_files,  # https://github.com/uiri/toml/issues/270
        'resources_dir': prj.resources_dir,
        **prj.metainfo,
        **prj.setuptools,
    }


def add_spec_to_pyproject_toml(pyproject_toml: Path, prj: BigflowProjectSpec):
    if pyproject_toml.exists():
        data = toml.load(pyproject_toml)
    else:
        data = {}
    data.setdefault('bigflow-project', {}).update(
        render_project_spec(prj)
    )
    data.setdefault('build-system', {
        'requires': [f"bigflow=={bigflow.__version__}"],
        'build-backend': "bigflow.build.meta",
    })
    pyproject_toml.write_text(toml.dumps(data))


@functools.lru_cache(maxsize=None)
def get_project_spec(project_dir: Path = None):
    """Reads project spec from `setup.py` and/or `pyproject.toml`.

    Memoize results (key = project path).  Intented for use from `bigflow.cli` and similar tools.
    """

    project_dir = project_dir or Path.cwd()
    return read_project_spec(project_dir)


def _maybe_read_pyproject(dir: Path):
    pyproject_toml = dir / "pyproject.toml"
    if pyproject_toml.exists():
        logger.info("Load config %s", pyproject_toml)
        data = toml.load(pyproject_toml)
        return data.get('bigflow-project')
    else:
        logger.debug("File %s not found", pyproject_toml)


def read_project_spec_nosetuppy(project_dir, **kwargs):
    """Read project spec from pyproject.toml, allowing to overwrite any options.
    Does *NOT* inspect `setup.py` (as it is intented to be used from setup.py)"""

    data = {}
    data.update(_maybe_read_pyproject(project_dir) or {})
    data.update(kwargs)
    return parse_project_spec(project_dir, **data)


def read_project_spec(dir: Path) -> BigflowProjectSpec:
    """Reads project spec from `setup.py` and/or `pyproject.toml`"""

    setuppy = dir / "setup.py"
    prj_setuppy = dir / "project_setup.py"  # legacy
    if not setuppy.exists() and prj_setuppy.exists():
        logger.info("Use deprecated 'project_setup.py' instead of 'setup.py'")
        setuppy = prj_setuppy

    if setuppy.exists():
        logger.debug("Read project spec from `setup.py` and `pyproject.toml`")
        setuppy_kwargs = bigflow.build.dev.read_setuppy_args(setuppy)
    else:
        logger.debug("Read project spec only from `pyproject.toml`")
        setuppy_kwargs = {}

    try:
        return read_project_spec_nosetuppy(dir, **setuppy_kwargs)
    except Exception as e:
        raise ValueError(
            "The project configuration is invalid. "
            "Check the documentation how to create a valid `setup.py`: "
            "https://github.com/allegro/bigflow/blob/master/docs/build.md"
        ) from e


# Provide defaults for project-spec

def discover_project_packages(project_dir: Path):
    ret = setuptools.find_packages(where=project_dir, exclude=["test.*", "test"])
    logger.info(
        "Automatically discovered %d packages: \n%s",
        len(ret),
        "\n".join(map(" - %s".__mod__, ret)),
    )
    return ret


def read_project_requirements(project_requirements_file):
    logger.info("Read project requirements from %s", project_requirements_file)
    req_txt = Path(project_requirements_file)
    recompiled = bigflow.build.pip.maybe_recompile_requirements_file(req_txt)
    if recompiled:
        logger.warning(textwrap.dedent(f"""
            !!! Requirements file was recompiled, you need to reinstall packages.
            !!! Run this command from your virtualenv:
            pip install -r {req_txt}
        """))
    bigflow.build.dataflow.dependency_checker.check_beam_worker_dependencies_conflict(req_txt)  # XXX
    return bigflow.build.pip.read_requirements(req_txt)


def get_docker_repository_from_deployment_config(deployment_config_file: Path) -> str:
    logger.info("Read docker repository from %s", deployment_config_file)

    import bigflow.cli   # TODO: refactor, remove this import
    try:
        config = bigflow.cli.import_deployment_config(str(deployment_config_file), 'docker_repository')
    except ValueError as e:
        raise ValueError(f"Can't find the specified deployment configuration: {deployment_config_file}") from e

    if isinstance(config, bigflow.Config):
        config = config.resolve()

    if "docker_repository" in config and not config["docker_repository"].islower():
        raise ValueError("`docker_repository` variable should be in lower case")
    docker_repository = config['docker_repository']

    if docker_repository is None:
        raise ValueError(f"Can't find the 'docker_repository' property in the specified config file: {deployment_config_file}")
    return docker_repository


def secure_get_version() -> str:
    logger.debug("Autodetected project version using git")
    try:
        version = bigflow.version.get_version()
        logger.info("Autodetected project version is %s", version)
        return version
    except Exception as e:
        logger.error("Can't get the current package version. To use the automatic versioning, "
                     "you need to use git inside your project directory: %s", e)
        return "INVALID"



import logging

from pathlib import Path

import bigflow
from bigflow.scaffold.templating import render_builtin_templates


logger = logging.getLogger(__name__)


def start_project(
    is_basic,
    project_name,
    projects_id,
    composers_bucket,
    envs,
    pyspark_job,
):
    project_dir = project_name + '_project'
    project_path = Path(project_dir).resolve()

    render_builtin_templates(
        project_path,
        "new-project",
        variables={
            'is_basic': is_basic,
            'project_id': projects_id[0],
            'dags_bucket': composers_bucket[0],
            'bigflow_version': bigflow.__version__,
            'project_name': project_name,
            'envs': envs,
            'composers_bucket': composers_bucket,
            pyspark_job: True,
        },
    )

    import bigflow.build.pip as bf_pip
    bf_pip.pip_compile(project_path / "resources" / "requirements.txt")


def migrate_project_from_10(
    project_path: Path,
    project_name,
):
    render_builtin_templates(project_path, 'migrate-11', variables={
        'project_name': project_name,
        'bigflow_version': bigflow.__version__,
    })



import logging
import unittest

from unittest import mock

from test import mixins

import bigflow.build.pip as bf_pip
import bigflow.build.dataflow.dependency_checker as dc


class DataflowDepsCheckTestCase(
    mixins.TempCwdMixin,
    mixins.FileUtilsMixin,
    unittest.TestCase,
):
    def setUp(self):
        super().setUp()
        self.req_in = self.cwd / "req.in"
        self.req_txt = self.cwd / "req.txt"
        self.pins_in = self.cwd / "dataflow_pins.in"

    @mock.patch('bigflow.build.dataflow.dependency_checker.detect_py_version')
    def test_detect_dataflow_conflicts(self, detect_py_version):
        # given
        detect_py_version.return_value = "3.7"
        self.req_in.write_text("""
            apache-beam==2.25
            freezegun==0.3.14
            fastavro==1.2.0
            requests==2.25.0
        """)
        bf_pip.pip_compile(self.req_in)

        # when
        conflicts = dc.detect_dataflow_conflicts(self.req_in)

        # then
        self.assertEqual(conflicts.get('freezegun'), ("0.3.14", "0.3.15"))
        self.assertEqual(conflicts.get('fastavro'), ("1.2.0", "0.24.2"))
        self.assertEqual(conflicts.get('requests'), ("2.25.0", "2.24.0"))

    def test_check_dataflow_conflicts(self):
        # given
        self.req_in.write_text("""
            apache-beam==2.25
            freezegun==0.3.14
        """)
        bf_pip.pip_compile(self.req_in)

        # then
        with self.assertLogs(dc.logger, level=logging.ERROR):
            dc.check_beam_worker_dependencies_conflict(self.req_in)

    @mock.patch('bigflow.build.dataflow.dependency_checker.detect_py_version')
    @mock.patch('bigflow.build.spec.get_project_spec')
    def test_generate_dataflow_pinfile(self, get_project_spec, detect_py_version):
        # given
        detect_py_version.return_value = "3.7"
        get_project_spec.return_value.project_requirements_file = str(self.req_txt)
        self.req_in.write_text("""
            # see `beam2.25_py3.7.txt`
            fastavro==0.24.2
            typing-extensions==3.7.4.2
            grpcio==1.30.0
            httplib2==0.17.3
            pymongo==3.10.1
            requests==2.24.0
            protobuf==3.12.2
            chardet==3.0.4
            pbr==5.5.0
            avro-python3==1.8.2
            urllib3==1.25.10
            certifi==2020.6.20
            oauth2client==3.0.0
            numpy==1.18.4
            #pytz==2020.1
            #rsa==4.6

            apache-beam==2.25
            rsa==4.7  # unersolvable conflict
            pytz      # latest version, resolvable conflict
        """)

        bf_pip.pip_compile(self.req_in)

        with self.assertLogs(level=logging.WARNING) as logs:
            dc.sync_requirements_with_dataflow_workers()

        # then
        self.assertFileExists(self.pins_in)
        self.assertFileContentRegex(self.req_in, r"(?mi)^-r dataflow_pins.in")

        self.assertFileContentRegex(self.pins_in, r"(?m)^## rsa==4.6")
        self.assertFileContentRegex(self.pins_in, r"(?m)^pytz==2020.1$")

        logs = "\n".join(logs.output)
        self.assertRegex(logs, r"Failed to pin some libraries")
        self.assertRegex(logs, r"\Wrsa\W")


import logging
import typing
import sys
import uuid
import os
import json

from textwrap import dedent

import google.cloud.logging
import google.cloud.logging.handlers

from urllib.parse import quote_plus
from bigflow.commons import public

try:
    from typing import TypedDict
except ImportError:
    TypedDict = dict


logger = logging.getLogger(__name__)


def create_gcp_log_handler(
    project_id,
    log_name,
    labels,
):
    client = google.cloud.logging.Client(project=project_id)
    handler = google.cloud.logging.handlers.CloudLoggingHandler(client, name=log_name, labels=labels)
    handler.addFilter(lambda r: not r.name.startswith("google.cloud.logging"))
    return handler


def prepare_gcp_logs_link(query: str):
    return f"https://console.cloud.google.com/logs/query;query={quote_plus(query)}"


def _uncaught_exception_handler(logger):
    def handler(exception_type, value, traceback):
        logger.error(f'Uncaught exception: {value}', exc_info=(exception_type, value, traceback))
    return handler


_LOGGING_CONFIGURED = False


# TODO: move to some shared config module?
class LogConfigDict(TypedDict):
    gcp_project_id: str
    log_name: str
    log_level: typing.Union[str, int]


def _generate_cl_log_view_query(params: dict):
    query = "".join(
        '{}"{}"\n'.format(k, v)
        for k, v in params.items()
    )
    return query


def workflow_logs_link_for_cli(log_config, workflow_id):
    log_name = log_config.get('log_name', workflow_id)
    full_log_name = f"projects/{log_config['gcp_project_id']}/logs/{log_name}"
    return prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.workflow_id=': workflow_id}))


def infrastructure_logs_link_for_cli(projects_config):
    links = {}
    for project_id, workflow_id in projects_config:
        links[project_id] = get_infrastructure_bigflow_project_logs(project_id)
    return links


def print_log_links_message(workflows_links, infra_links):
    workflows_links = '\n'.join(f'{workflow}: {link}' for workflow, link in workflows_links.items())
    infra_links = '\n'.join(f'{workflow}: {link}' for workflow, link in infra_links.items())
    print(dedent(f"""
*************************LOGS LINK*************************
Infrastructure logs:
{infra_links}
Workflow logs:
{workflows_links}
***********************************************************"""))


def get_infrastructure_bigflow_project_logs(project_id):
    pod_errors = _generate_cl_log_view_query({"severity>=": "INFO"}) + _generate_cl_log_view_query(
        {"resource.type=": "k8s_pod"}) + '"Error:"'
    dataflow_errors = _generate_cl_log_view_query({"resource.type=": "dataflow_step",
                                                   "log_name=": f"projects/{project_id}/logs/dataflow.googleapis.com%2Fjob-message"}) + _generate_cl_log_view_query(
        {"severity>=": "INFO"})

    result = []
    for entry in [pod_errors, dataflow_errors]:
        result.append(f'({entry})')

    condition = '\nOR\n'.join(result)
    return prepare_gcp_logs_link(condition)


def init_workflow_logging(workflow: 'bigflow.Workflow', banner=True):
    if workflow.log_config:
        init_logging(workflow.log_config, workflow.workflow_id, banner=banner)
    else:
        print("Log config is not provided for the Workflow")


@public()
def init_logging(config: LogConfigDict, workflow_id: str, banner=True):

    global _LOGGING_CONFIGURED
    if _LOGGING_CONFIGURED:
        import warnings
        warnings.warn(UserWarning("bigflow.log is already configured - skip"))
        return
    _LOGGING_CONFIGURED = True

    gcp_project_id = config['gcp_project_id']
    log_name = config.get('log_name', workflow_id)
    log_level = config.get('log_level', 'INFO')
    run_uuid = str(uuid.uuid4())

    labels = {
        'workflow_id': workflow_id,
        'run_uuid': run_uuid,
    }

    root = logging.getLogger()
    if not root.handlers:
        # logs are not configured yet - print to stderr
        logging.basicConfig(level=log_level)
    elif log_level:
        root.setLevel(min(root.level, logging._checkLevel(log_level)))

    full_log_name = f"projects/{gcp_project_id}/logs/{log_name}"
    infrastructure_logs = get_infrastructure_bigflow_project_logs(gcp_project_id)
    workflow_logs_link = prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.workflow_id=': workflow_id}))
    this_execution_logs_link = prepare_gcp_logs_link(
        _generate_cl_log_view_query({'logName=': full_log_name, 'labels.run_uuid=': run_uuid}))

    if banner:
        logger.info(dedent(f"""
               *************************LOGS LINK*************************
               Infrastructure logs:{infrastructure_logs}
               Workflow logs (all runs): {workflow_logs_link}
               Only this run logs: {this_execution_logs_link}
               ***********************************************************"""))
    gcp_logger_handler = create_gcp_log_handler(gcp_project_id, log_name, labels)
    gcp_logger_handler.setLevel(log_level or logging.INFO)

    # Disable logs from 'google.cloud.logging'
    gclogging_logger = logging.getLogger("google.cloud.logging")
    gclogging_logger.setLevel(logging.WARNING)
    gclogging_logger.propagate = False
    gclogging_logger.addHandler(logging.StreamHandler())

    # TODO: add formatter?
    root.addHandler(gcp_logger_handler)

    sys.excepthook = _uncaught_exception_handler(logging.getLogger('uncaught_exception'))


def maybe_init_logging_from_env():
    if 'bf_log_config' not in os.environ:
        return

    log_config = os.environ.get('bf_log_config', "{}")
    try:
        log_config = json.loads(log_config)
    except ValueError as e:
        print("invalid 'log_config' json:", e, file=sys.stderr)
        return

    if 'workflow_id' in log_config:
        workflow_id = log_config['workflow_id']
    else:
        workflow_id = os.environ.get('bf_workflow_id')

    init_logging(log_config, workflow_id or 'none', banner=False)



import os
import re
import unittest
import pickle

from test import mixins


class SelfBuildOldProjectTestCase(
    mixins.SubprocessMixin,
    mixins.PrototypedDirMixin,
    mixins.BigflowInPythonPathMixin,
    unittest.TestCase,
):
    proto_dir = "bf-projects/bf_simple_v10"

    def test_should_build_bdist(self):

        # given
        self.assertFileNotExists("dist/*.whl")
        self.subprocess_run(["python", "project_setup.py", "bdist_wheel"])

        # then
        self.assertFileExists("dist/*.whl")


class _BaseBuildReflectTest(
    mixins.SubprocessMixin,
    mixins.PrototypedDirMixin,
    unittest.TestCase,
):
    proto_dir = "buildd/bf-projects/bf_selfbuild_project"

    def runpy_n_dump(self, func_name: str):
        mod, _ = func_name.rsplit(".", 1)
        pycode = f"import {mod}, pickle, os; pickle.dump({func_name}(), os.fdopen(1, 'wb'))"
        r = self.subprocess_run(["python", "-c", pycode], check=True, capture_output=True)
        return pickle.loads(r.stdout)

    def check_build_reflect(self):

        # then - check reading of project spec
        self.assertEqual("bf-selfbuild-project", self.runpy_n_dump('bf_selfbuild_project.buildme.project_spec').name)
        self.assertEqual("bf-selfbuild-project", self.runpy_n_dump('bf_selfbuild_other_package.buildme.project_spec').name)
        #self.assertEqual("bf-selfbuild-project", self.runpy_n_dump('bf_selfbuild_module.project_spec').name)

        # then - self-build sdist/wheel/egg pacakges
        sdist_pkg = self.runpy_n_dump('bf_selfbuild_project.buildme.build_sdist')
        self.addCleanup(os.unlink, sdist_pkg)
        self.assertRegex(str(sdist_pkg), r".*\.tar\.gz")
        self.assertFileExists(sdist_pkg)

        wheel_pkg = self.runpy_n_dump('bf_selfbuild_project.buildme.build_wheel')
        self.addCleanup(os.unlink, wheel_pkg)
        self.assertRegex(str(wheel_pkg), r".*\.whl")
        self.assertFileExists(wheel_pkg)

        egg_pkg = self.runpy_n_dump('bf_selfbuild_project.buildme.build_egg')
        self.addCleanup(os.unlink, egg_pkg)
        self.assertRegex(str(egg_pkg), r".*\.egg")
        self.assertFileExists(egg_pkg)

        # then - verify materizlied setuppy
        setuppy = self.runpy_n_dump('bf_selfbuild_project.buildme.materialize_setuppy')
        self.addCleanup(os.unlink, setuppy)
        self.assertFileContentRegex(setuppy, r"import bigflow.build")
        self.assertFileContentRegex(setuppy, re.compile(r"bigflow.build.setup(.*)", re.M))


class SelfBuildProjectFromPackageTestCase(
    mixins.VenvMixin,
    _BaseBuildReflectTest,
):
    def test_reflected_build_from_wheel(self):

        # build 'whl' package
        self.assertFileNotExists("dist/*.whl")
        self.subprocess_run(["python", "setup.py", "bdist_wheel"])

        # install .whl package
        whl = self.assertFileExists("dist/*.whl")
        self.venv_pip_install(whl)

        # remove original sdist - drop cwd & create a new one
        self.chdir_new_temp()

        # then
        self.check_build_reflect()


class SelfBuildProjectFromSourcesTestCase(
    _BaseBuildReflectTest,
    mixins.BigflowInPythonPathMixin,
):
    def test_reflected_build_from_sources(self):
        # then
        self.check_build_reflect()

        # and - test pass (check if spec is available from tests)
        self.subprocess_run(["python", "test/test_readspec.py"])

        # and - project spec is available from scripts
        self.subprocess_run(["python", "scripts/subdir/infer_project_name.py"])
